#!/usr/bin/env python
# coding: utf-8
"""Tests for the mesh import of mesh from external libraries"""

import os
import pytest
import logging
from unittest import mock
from pathlib import Path

import numpy as np
import xarray as xr

from capytaine.io.mesh_writers import write_STL, write_GDF
from capytaine.io.mesh_loaders import load_STL, load_HST, load_GDF
import capytaine as cpt

try:
    import meshio
    import pygmsh
    import gmsh
except ImportError:
    pygmsh = None

try:
    import h5py
except ImportError:
    h5py = None

offset=1e-2 # small offset so you can clip at z=0
omega = 0.5*2*np.pi
rho_water = 1e3

#################################################################

def parse_hst(file_content):
    # Workaround to avoid actually writing/reading a file on disk
    with mock.patch('capytaine.io.mesh_loaders._check_file', lambda foo: None):
        with mock.patch('capytaine.io.mesh_loaders.open', mock.mock_open(read_data=file_content)):
            return load_HST("mocked/filename")


def test_hst_type_0_triangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[1/2, 1/2, 0.0]]))


def test_hst_type_0_quadrangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[2/3, 2/3, 0.0]]))


def test_hst_type_1_quadrangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 1
        1 1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[1/2, 1/2, 0.0]]))


def test_hst_type_1_triangle():
    mesh = parse_hst("""
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 1
        1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[2/3, 2/3, 0.0]]))


def test_hst_implicit_coordinate_numbering():
    mesh = parse_hst("""
        COORDINATES
        0.0 0.0 0.0
        1.0 0.0 0.0
        1.0 1.0 0.0
        0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
        """)
    assert mesh.vertices.shape == (4, 3)
    assert mesh.faces.shape == (1, 4)
    assert np.allclose(mesh.faces_centers, np.array([[1/2, 1/2, 0.0]]))


def test_hst_coordinate_numbering_error():
    with pytest.raises(ValueError):
        mesh = parse_hst("""
            COORDINATES
            1 0.0 0.0 0.0
            2 1.0 0.0 0.0
            3 1.0 1.0 0.0
            5 0.0 1.0 0.0
            ENDCOORDINATES
            """)

def test_hst_symmetry_1():
    mesh = parse_hst("""
        SYMMETRY 1
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half, cpt.Mesh)
    assert mesh.merged().nb_faces == 2


def test_hst_symmetry_2():
    mesh = parse_hst("""
        SYMMETRY 2
        COORDINATES
        1 0.0 0.0 0.0
        2 1.0 0.0 0.0
        3 1.0 1.0 0.0
        4 0.0 1.0 0.0
        ENDCOORDINATES
        PANEL TYPE 0
        1 2 3 4
        ENDPANEL
            """)
    assert isinstance(mesh, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half, cpt.ReflectionSymmetricMesh)
    assert isinstance(mesh.half.half, cpt.Mesh)
    assert mesh.merged().nb_faces == 4


def test_hst_ignored_lines(caplog):
    with caplog.at_level(logging.WARNING):
        mesh = parse_hst("""
            USER mancellin
            NBBODY 1
            COORDINATES
            1 0.0 0.0 0.0
            2 1.0 0.0 0.0
            3 1.0 1.0 0.0
            4 0.0 1.0 0.0
            ENDCOORDINATES
            PANEL TYPE 0
            1 2 3 4
            ENDPANEL
                """)
    assert "HST mesh reader ignored" in caplog.text
    assert "USER mancellin" in caplog.text
    assert "NBBODY 1" in caplog.text

#################################################################

def generate_pygmsh_wavebot():
    T1=0.16
    T2=0.37
    r1=0.88
    r2=0.35
    with pygmsh.occ.Geometry() as geom:
        cyl = geom.add_cylinder([0,0,0], [0,0,-T1], r1)
        cone = geom.add_cone([0,0,-T1], [0,0,-T2], r1, r2)
        geom.translate(cyl, [0, 0, offset])
        geom.translate(cone, [0, 0, offset])
        geom.boolean_union([cyl, cone])
        mesh = geom.generate_mesh(dim=2)
    vol_exp = np.pi*T1*r1**2 + 1/3 * np.pi * T2 * (r2**2 + r2 * r1 + r1**2)
    return (mesh, vol_exp)

def generate_pygmsh_cylinder():
    T=0.52
    r1=0.88
    with pygmsh.occ.Geometry() as geom:
        cyl = geom.add_cylinder([0,0,0], [0,0,-T], r1)
        geom.translate(cyl, [0, 0, offset])
        mesh = geom.generate_mesh(dim=2)
    vol_exp = np.pi*r1**2*T
    return (mesh, vol_exp)

def generate_pygmsh_sphere():
    r1=0.88
    with pygmsh.occ.Geometry() as geom:
        sphere = geom.add_ball(center=[0,0,0], radius=r1)
        geom.translate(sphere, [0, 0, offset])
        mesh = geom.generate_mesh(dim=2)
    vol_exp = 1/2*np.pi*4/3*r1**3
    return (mesh, vol_exp)


@pytest.mark.skipif(pygmsh is None,
                    reason='pygmsh and/or meshio is not installed')
@pytest.mark.parametrize("generate_pygmsh",
                         [generate_pygmsh_cylinder, generate_pygmsh_sphere, generate_pygmsh_wavebot])
def test_from_meshio_pygmsh(generate_pygmsh, tmp_path):
    mesh, vol_exp = generate_pygmsh()
    fb = cpt.FloatingBody(mesh=mesh)
    fb.mesh.heal_mesh()
    fb.keep_immersed_part()

    vol = fb.mesh.volume
    assert pytest.approx(vol_exp, rel=1e-1) == vol

    fb.add_translation_dof((0,0,1))

    test_matrix = xr.Dataset(coords={
        'rho': rho_water,
        'water_depth': [np.inf],
        'omega': omega,
        'wave_direction': 0,
        'radiating_dof': list(fb.dofs.keys()),
        })

    solver = cpt.BEMSolver()
    data = solver.fill_dataset(test_matrix,
                               bodies=[fb],
                               hydrostatics=True,
                               mesh=True,
                               wavelength=True,
                               wavenumber=True)

#################################################################

def test_STL(tmp_path):
    vtk = pytest.importorskip("vtk", reason="VTK not installed, test skipped.")
    mesh = cpt.mesh_sphere()
    filepath = tmp_path / "test.stl"
    write_STL(filepath, mesh.vertices, mesh.faces)
    reloaded_mesh = load_STL(str(filepath), name="Bla")

    assert reloaded_mesh.name == "Bla"
    assert np.allclose(mesh.vertices, reloaded_mesh.vertices)

def test_STL_with_uppercase_file_extension(tmp_path):
    vtk = pytest.importorskip("vtk", reason="VTK not installed, test skipped.")
    mesh = cpt.mesh_sphere()
    filepath = tmp_path / "test.STL"
    write_STL(filepath, mesh.vertices, mesh.faces)
    reloaded_mesh = cpt.load_mesh(str(filepath), name="Bla")
    # Should detect that the file is stl using the file extension

    assert reloaded_mesh.name == "Bla"
    assert np.allclose(mesh.vertices, reloaded_mesh.vertices)

@pytest.mark.skipif(pygmsh is None,
                    reason='pygmsh and/or meshio is not installed')
def test_STL_with_meshio(tmp_path):
    vtk = pytest.importorskip("vtk", reason="VTK not installed, test skipped.")
    mesh, _ = generate_pygmsh_cylinder()
    from capytaine.io.meshio import load_from_meshio
    cpt_mesh = load_from_meshio(mesh)
    cpt_mesh.heal_mesh()

    # Write with Meshio and reload with Meshmagick
    mesh.write(tmp_path / "wavebot.stl")
    cpt_mesh_2 = cpt.load_mesh(tmp_path / "wavebot.stl")
    assert np.allclose(cpt_mesh.vertices, cpt_mesh_2.vertices)

    # # Write with Meshmagick and reload with Meshio
    # # FAILING FOR NOW
    # # ERROR IN STL LOADER IN MESHMAGICK
    # write_STL(tmp_path / "wavebot2.stl", fb.mesh.vertices, fb.mesh.faces)
    # mesh2 = meshio.read(tmp_path / "wavebot2.stl")
    # fb3 = cpt.FloatingBody.from_meshio(mesh2)
    # assert np.allclose(fb.mesh.vertices, fb3.mesh.vertices)

#################################################################

def test_write_and_load_gdf(tmpdir):
    mesh_path = tmpdir.join("temp_mesh.gdf")

    original_mesh = cpt.mesh_horizontal_cylinder()
    write_GDF(str(mesh_path), original_mesh.vertices, original_mesh.faces, ulen=1, gravity=9.81, isx=0, isy=0)

    read_mesh = load_GDF(str(mesh_path))

    np.testing.assert_allclose(
        read_mesh.vertices[read_mesh.faces],
        original_mesh.vertices[original_mesh.faces],
        atol=1e-6
        )

    np.testing.assert_allclose(
        read_mesh.faces_areas,
        original_mesh.faces_areas
        )

    np.testing.assert_allclose(
        read_mesh.faces_normals,
        original_mesh.faces_normals,
        atol=1e-6
        )

    np.testing.assert_allclose(
        read_mesh.faces_centers,
        original_mesh.faces_centers
        )

    np.testing.assert_allclose(
        read_mesh.faces_radiuses,
        original_mesh.faces_radiuses
        )

    np.testing.assert_allclose(
        read_mesh.volume,
        original_mesh.volume
        )


@pytest.mark.skipif(h5py is None,
                    reason='h5py and/or meshio is not installed')
def test_MED_file():
    mesh = cpt.load_mesh(os.path.join(os.path.dirname(__file__), "mesh_files_examples/barge.med"))
    assert mesh.nb_faces == 187


def test_pnl(tmpdir):
    mesh_path = tmpdir.join("temp_mesh.pnl")
    mesh = cpt.mesh_sphere()
    from capytaine.io.mesh_writers import write_PNL
    write_PNL(mesh_path, mesh.vertices, mesh.faces)
    reloaded_mesh = cpt.load_mesh(mesh_path)
    np.testing.assert_equal(mesh.faces, reloaded_mesh.faces)
    np.testing.assert_allclose(mesh.vertices, reloaded_mesh.vertices, atol=1e-5)



#!/usr/bin/env python
# coding: utf-8
"""
Tests for the functions that computes hydrostatic from the mesh vertices
and faces
"""

import logging
import pytest
import json
from pathlib import Path

import pandas as pd
import numpy as np
import capytaine as cpt

# TODO: Can I use pytest fixtures to avoid regenerating so many spheres?
# Does pytest fixture do a copy of the object, since I modifying the sphere in-place?

def test_mesh_properties():
    sphere = cpt.Sphere(radius=1.0, center=(0, 0, -2), nphi=50, ntheta=50)
    assert np.allclose(sphere.center_of_buoyancy, sphere.mesh.center_of_buoyancy)

def test_disp_mass_of_sphere():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,-2), nphi=50, ntheta=50)
    analytical_volume = 4/3*np.pi*1.0**3
    assert np.isclose(sphere.disp_mass(rho=1000), 1000*analytical_volume, rtol=2e-2)

def test_waterplane_area_of_submerged_sphere():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,-2), nphi=20, ntheta=20)
    assert np.isclose(sphere.waterplane_area, 0.0)

def test_waterplane_center_of_submerged_sphere():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,-2), nphi=20, ntheta=20)
    assert sphere.waterplane_center is None

def test_waterplane_center_of_sphere_at_surface():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    assert np.allclose(sphere.waterplane_center, [0.0, 0.0])

def test_infer_rotation_center():
    sphere = cpt.Sphere().keep_immersed_part()
    sphere.rotation_center = (7, 8, 9)
    sphere.add_all_rigid_body_dofs()
    assert np.allclose(sphere._infer_rotation_center(), (7, 8, 9))
    del sphere.rotation_center
    assert np.allclose(sphere._infer_rotation_center(), (7, 8, 9))

def test_stiffness_when_no_dofs():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.center_of_mass = np.array([0, 0, -0.3])
    with pytest.raises(AttributeError, match=".* no dof .*"):
        sphere.compute_hydrostatic_stiffness()

def test_stiffness_no_cog():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.add_all_rigid_body_dofs()
    with pytest.raises(ValueError, match=".*no center of mass.*"):
        sphere.compute_hydrostatic_stiffness()

def test_stiffness_dof_ordering():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20)
    sphere.keep_immersed_part()
    sphere.add_all_rigid_body_dofs()
    sphere.center_of_mass = np.array([0.0, 0.0, -0.2])
    K = sphere.compute_hydrostatic_stiffness()
    assert np.all(K.coords["radiating_dof"].values == np.array(['Surge', 'Sway', 'Heave', 'Roll', 'Pitch', 'Yaw']))

def test_stifness_rigid_body_invariance_by_translation():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20)
    sphere.keep_immersed_part()
    sphere.center_of_mass = np.array([0.0, 0.0, -0.2])
    sphere.add_all_rigid_body_dofs()
    K1 = sphere.compute_hydrostatic_stiffness()
    K2 = sphere.translated([1.0, 0.0, 0.0]).compute_hydrostatic_stiffness()
    assert np.allclose(K1, K2)

def test_stifness_generalized_dof_invariance_by_translation():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20)
    sphere.keep_immersed_part()
    # Not really a generalized dof, but since the dofs have non-standard names
    # the code will not recognize the rigid body dof.
    sphere.add_translation_dof(direction=(1, 0, 0), name="cavalement")
    sphere.add_rotation_dof(axis=cpt.Axis(vector=(0, 1, 0), point=(0, 0, 0)), name="tangage")
    K1 = sphere.compute_hydrostatic_stiffness()
    K2 = sphere.translated([1.0, 0.0, 0.0]).compute_hydrostatic_stiffness()
    assert np.allclose(K1, K2)

def test_stiffness_with_divergence():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.center_of_mass = np.array([0, 0, -0.3])
    sphere.dofs["elongate_in_z"] = np.array([(0, 0, z) for (x, y, z) in sphere.mesh.faces_centers])
    hs_1 = sphere.compute_hydrostatic_stiffness()
    hs_2 = sphere.compute_hydrostatic_stiffness(divergence={"elongate_in_z": np.ones(sphere.mesh.nb_faces)})
    assert hs_1.values[0, 0] != hs_2.values[0, 0]
    analytical_hs = - 1000.0 * 9.81 * (4 * sphere.volume * sphere.center_of_buoyancy[2])
    assert np.isclose(hs_2.values[0, 0], analytical_hs)

def test_stiffness_with_malformed_divergence(caplog):
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.center_of_mass = np.array([0, 0, -0.3])
    sphere.dofs["elongate_in_z"] = np.array([(0, 0, z) for (x, y, z) in sphere.mesh.faces_centers])
    hs_1 = sphere.compute_hydrostatic_stiffness()
    with caplog.at_level(logging.WARNING):
        hs_2 = sphere.compute_hydrostatic_stiffness(divergence={"foobar": np.ones(sphere.mesh.nb_faces)})
    assert hs_1.values[0, 0] == hs_2.values[0, 0]
    assert "without the divergence" in caplog.text

def test_stiffness_joined_bodies():
    a = cpt.Sphere(name="foo")
    a.add_translation_dof(name="Heave")
    a.hydrostatic_stiffness = a.add_dofs_labels_to_matrix(np.array([[1.0]]))
    b = cpt.RectangularParallelepiped(name="bar")
    b.add_all_rigid_body_dofs()
    b.hydrostatic_stiffness = b.add_dofs_labels_to_matrix(np.random.rand(6, 6))
    both = a + b
    assert both.hydrostatic_stiffness.shape == (7, 7)

def test_mass_of_sphere_for_non_default_density():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,-2), nphi=50, ntheta=50)
    sphere.add_all_rigid_body_dofs()
    sphere.center_of_mass = np.array([0, 0, -2])
    m = sphere.compute_rigid_body_inertia(rho=500)
    analytical_volume = 4/3*np.pi*1.0**3
    assert np.isclose(m.values[0, 0], 500*analytical_volume, rtol=2e-2)

def test_inertia_rigid_body_dofs():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.center_of_mass = np.array([0, 0, -0.3])
    sphere.add_all_rigid_body_dofs()
    assert np.all(sphere.compute_rigid_body_inertia(output_type="rigid_dofs")
            == sphere.compute_rigid_body_inertia(output_type="all_dofs"))
    assert np.all(sphere.compute_rigid_body_inertia(output_type="rigid_dofs")
            == sphere.compute_rigid_body_inertia(output_type="body_dofs"))

def test_inertia_invariance_by_translation():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20)
    sphere.keep_immersed_part()
    sphere.center_of_mass = np.array([0.0, 0.0, -0.2])
    sphere.add_all_rigid_body_dofs()
    M1 = sphere.compute_rigid_body_inertia()
    M2 = sphere.translated([1.0, 0.0, 0.0]).compute_rigid_body_inertia()
    assert np.allclose(M1, M2)

def test_inertia_wrong_output_type():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.center_of_mass = np.array([0, 0, -0.3])
    sphere.add_all_rigid_body_dofs()
    with pytest.raises(ValueError):
        sphere.compute_rigid_body_inertia(output_type="foo")

def test_inertia_when_no_dofs():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.rotation_center = np.array([0, 0, 0])
    sphere.center_of_mass = np.array([0, 0, -0.3])
    m =sphere.compute_rigid_body_inertia()
    assert m.shape == (0, 0)

def test_inertia_no_cog():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20).keep_immersed_part()
    sphere.add_all_rigid_body_dofs()
    with pytest.raises(ValueError, match=".*no center of mass.*"):
        sphere.compute_rigid_body_inertia()

def test_inertia_joined_bodies():
    a = cpt.Sphere(name="foo")
    a.add_translation_dof(name="Heave")
    a.inertia_matrix = a.add_dofs_labels_to_matrix(np.array([[1.0]]))
    b = cpt.RectangularParallelepiped(name="bar")
    b.add_all_rigid_body_dofs()
    b.inertia_matrix = b.add_dofs_labels_to_matrix(np.random.rand(6, 6))
    both = a + b
    assert both.inertia_matrix.shape == (7, 7)

def test_inertia_joined_bodies_with_missing_inertia():
    a = cpt.Sphere(name="foo")
    a.add_translation_dof(name="Heave")
    # No inertia matrix
    b = cpt.RectangularParallelepiped(name="bar")
    b.add_all_rigid_body_dofs()
    b.inertia_matrix = b.add_dofs_labels_to_matrix(np.random.rand(6, 6))
    both = a + b
    assert not hasattr(both, "inertia_matrix")

def test_inertia_joined_bodies_associativity():
    a = cpt.Sphere(name="foo")
    a.add_translation_dof(name="Heave")
    a.inertia_matrix = a.add_dofs_labels_to_matrix(np.array([[1.0]]))
    b = cpt.RectangularParallelepiped(name="bar")
    b.add_all_rigid_body_dofs()
    b.inertia_matrix = b.add_dofs_labels_to_matrix(np.random.rand(6, 6))
    c = cpt.VerticalCylinder(name="baz")
    c.add_rotation_dof(name="Pitch")
    c.inertia_matrix = c.add_dofs_labels_to_matrix(np.array([[2.0]]))
    i1 = ((a + b) + c).inertia_matrix
    i2 = (a + (b + c)).inertia_matrix
    i3 = cpt.FloatingBody.join_bodies(a, b, c).inertia_matrix
    assert np.allclose(i1.values, i2.values, i3.values)

def test_inertia_joined_bodies_commutativity():
    a = cpt.Sphere(name="foo")
    a.add_translation_dof(name="Heave")
    a.inertia_matrix = a.add_dofs_labels_to_matrix(np.array([[1.0]]))
    b = cpt.RectangularParallelepiped(name="bar")
    b.add_all_rigid_body_dofs()
    b.inertia_matrix = b.add_dofs_labels_to_matrix(np.random.rand(6, 6))
    i1 = (a + b).inertia_matrix
    i2 = (b + a).inertia_matrix
    assert not np.allclose(i1.values, i2.values)
    # but their are the same when the coordinates are sorted in the same way:
    assert np.allclose(i1.sel(influenced_dof=i2.influenced_dof, radiating_dof=i2.radiating_dof).values, i2.values)


def test_hydrostatics_of_submerged_sphere():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,-2), nphi=20, ntheta=20)
    sphere.add_all_rigid_body_dofs()
    sphere.center_of_mass = np.array([0, 0, -2])
    sphere.compute_hydrostatics()
    assert sphere.inertia_matrix.shape == (6, 6)

def test_hydrostatics_no_rigid_dof():
    mesh = cpt.mesh_sphere()
    body = cpt.FloatingBody(mesh=mesh, dofs={"Bulge": mesh.faces_normals}, center_of_mass=(0, 0, 0))
    hs = body.compute_hydrostatics()
    assert 'inertia_matrix' not in hs

def test_all_hydrostatics():
    density = 1000
    gravity = 9.80665

    sphere = cpt.Sphere(
        radius=10.0,
        center=(0,0,-1),
        nphi=100, ntheta=50,
    )
    horizontal_cylinder = cpt.HorizontalCylinder(
        length=10.0, radius=5.0,
        center=(0,10,-1),
        nr=100, nx=100, ntheta=10,
    )
    vertical_cylinder = cpt.VerticalCylinder(
        length=10.0, radius=5.0,
        center=(10,0,0),
        nr=200, nx=200, ntheta=10,
    )
    body = sphere + horizontal_cylinder + vertical_cylinder
    body.rotation_center = np.array([0, 0, 0])
    body.center_of_mass = body.center_of_buoyancy
    body.add_all_rigid_body_dofs()

    capy_hsdb = body.compute_hydrostatics(rho=density, g=gravity)

    stiff_compare_dofs = ["Heave", "Roll", "Pitch"]
    capy_hsdb["stiffness_matrix"] = capy_hsdb["hydrostatic_stiffness"].sel(
        influenced_dof=stiff_compare_dofs, radiating_dof=stiff_compare_dofs
        ).values

    mass_compare_dofs = ["Roll", "Pitch", "Yaw"]
    capy_hsdb["inertia_matrix"] = capy_hsdb["inertia_matrix"].sel(
        influenced_dof=mass_compare_dofs, radiating_dof=mass_compare_dofs
        ).values


    # =============================================================================
    # Meshmagick
    # =============================================================================
    case_dir = Path(__file__).parent / "Hydrostatics_cases"
    # case_dir.mkdir(parents=True, exist_ok=True)
    # import meshmagick.mesh as mmm
    # import meshmagick.hydrostatics as mmhs
    # body_mesh = mmm.Mesh(body.mesh.vertices, body.mesh.faces, name=body.mesh.name)
    # mm_hsdb = mmhs.compute_hydrostatics(body_mesh, body.center_of_mass, density, gravity)
    # mm_hsdb["inertia_matrix"] = body_mesh.eval_plain_mesh_inertias(
    # rho_medium=density).inertia_matrix
    # mm_hsdb["mesh"] = ""
    # with open(f'{case_dir}/sphere__hor_cyl__ver_cyl.pkl.json', 'w') as convert_file:
    #     mm_hsdb_json = {key:(value.tolist() if type(value)==np.ndarray else value)
    #                         for key, value in mm_hsdb.items() }
    #     convert_file.write(json.dumps(mm_hsdb_json))

    with open(f'{case_dir}/sphere__hor_cyl__ver_cyl.pkl.json', 'r',
              encoding="UTF-8") as f:
        mm_hsdb = json.load(f)

    # =============================================================================
    # Testing
    # =============================================================================

    for var in capy_hsdb.keys():
        if var in mm_hsdb.keys():
            # if not np.isclose(capy_hsdb[var], mm_hsdb[var],
            #                   rtol=1e-2, atol=1e-3).all():
            #     print(f"{var}:")
            #     print(f"    Capytaine  - {capy_hsdb[var]}")
            #     print(f"    Meshmagick - {mm_hsdb[var]}")
            assert np.isclose(capy_hsdb[var], mm_hsdb[var],
                              rtol=1e-2, atol=1e-3).all()


def test_vertical_elastic_dof():

    bodies = [
        cpt.VerticalCylinder(
            length=10.0, radius=5.0,
            center=(0.0,0.0,0.0),
            nr=100, nx=100, ntheta=100,
        ),

        cpt.Sphere(
            radius=100,
            center=(0,0,0),
            nphi=50, ntheta=50,
        ),

        cpt.HorizontalCylinder(
            length=5.0, radius=1.0,
            center=(0,10,0),
            reflection_symmetry=False,
            nr=20, nx=20, ntheta=10,
        )
    ]

    for body in bodies:
        body.center_of_mass = body.center_of_buoyancy
        body.keep_immersed_part()
        faces_centers = body.mesh.faces_centers
        body.dofs["elongate"] = np.zeros_like(faces_centers)
        body.dofs["elongate"][:,2] = faces_centers[:,2]

        divergence = np.ones(body.mesh.faces_centers.shape[0])

        density = 1000
        gravity = 9.80665
        capy_hs = body.each_hydrostatic_stiffness("elongate", "elongate",
                    influenced_dof_div=divergence, rho=density, g=gravity).values[0][0]

        analytical_hs = - (density * gravity * 4 * body.volume
                        * body.center_of_buoyancy[2])

        assert np.isclose(capy_hs, analytical_hs)


###########################
#  Non-neutrally buoyant  #
###########################

def test_mass_joined_bodies():
    a = cpt.FloatingBody(mass=100)
    b = cpt.FloatingBody(mass=300)
    assert (a + b).mass == 400

def test_mass_joined_bodies_with_missing_mass():
    a = cpt.FloatingBody()
    b = cpt.FloatingBody(mass=300)
    assert (a + b).mass is None

def test_center_of_mass_joined_bodies():
    a = cpt.FloatingBody(mass=100, center_of_mass=(0, 0, 0))
    b = cpt.FloatingBody(mass=300, center_of_mass=(1, 0, 0))
    assert np.allclose((a + b).center_of_mass, (0.75, 0, 0))

def test_center_of_mass_joined_bodies_with_missing_mass():
    a = cpt.FloatingBody()
    b = cpt.FloatingBody(mass=300, center_of_mass=(1, 0, 0))
    assert (a + b).center_of_mass is None

def test_not_single_rigid_and_non_neutrally_buoyant_body():
    m = cpt.mesh_sphere()
    a = cpt.FloatingBody(
        mesh=m, mass=100, center_of_mass=(0, 0, 0),
        dofs=cpt.rigid_body_dofs(rotation_center=(0, 0, 0)),
    )
    b = cpt.FloatingBody(
        mesh=m.translated_x(2.0), mass=300, center_of_mass=(2, 0, 0),
        dofs=cpt.rigid_body_dofs(rotation_center=(2, 0, 0)),
    )
    with pytest.raises(NotImplementedError):
        (a + b).compute_hydrostatic_stiffness()

def test_non_neutrally_buoyant_stiffness():
    body = cpt.VerticalCylinder(radius=1.0, length=2.0, center=(0.0, 0.0, 0.0), nx=40, ntheta=40, nr=20)
    body.rotation_center = (0, 0, 0)
    body.add_all_rigid_body_dofs()
    body.keep_immersed_part()
    body.mass = 500 * body.volume
    body.center_of_mass = (0.0, 0.0, -0.25)

    K = body.compute_hydrostatic_stiffness()
    dofs = ["Heave", "Roll", "Pitch", "Yaw"]
    K = K.sel(influenced_dof=dofs, radiating_dof=dofs).values
    # K is now a 4x4 np.array in correct order

    rho_g = 1000*9.81
    K_ref = rho_g * np.array([
        [3.137, -0.382e-3, -0.613e-4, 0.0       ],
        [0.0,   -0.392,    -0.276e-4, -0.448e-4 ],
        [0.0,   0.0,       -0.392,    0.313e-3  ],
        [0.0,   0.0,       0.0,       0.0       ],
        ])  # Computed with WAMIT
    print(K)
    print(K_ref)
    assert np.allclose(K, K_ref, atol=rho_g*1e-2)


def test_non_neutrally_buoyant_K55():
    body = cpt.VerticalCylinder(radius=1.0, length=2.0, center=(0.0, 0.0, 0.0), nx=40, ntheta=40, nr=20)
    body.rotation_center = (0, 0, 0)
    body.add_all_rigid_body_dofs()
    body.keep_immersed_part()

    ref_data = pd.DataFrame([
        dict(body_density=1000, z_cog=-0.00, K55=-7.70e3),
        dict(body_density=1000, z_cog=-0.25, K55=5.0),
        dict(body_density=1000, z_cog=-0.50, K55=7.67e3),
        dict(body_density=500,  z_cog=-0.00, K55=-7.67e3),
        dict(body_density=500,  z_cog=-0.25, K55=-3.83e3),
        dict(body_density=500,  z_cog=-0.50, K55=5.0),
        ])
    ref_data['mass'] = body.volume * ref_data['body_density']

    rho_g = 1000*9.81
    for (i, case) in ref_data.iterrows():
        body.mass = case.mass
        body.center_of_mass = (0.0, 0.0, case.z_cog)
        K55 = body.compute_hydrostatic_stiffness().sel(influenced_dof="Pitch", radiating_dof="Pitch").values
        assert np.isclose(K55, case.K55, atol=rho_g*1e-2)


def test_non_neutrally_buoyant_stiffness_invariance_by_translation():
    sphere = cpt.Sphere(radius=1.0, center=(0,0,0), nphi=20, ntheta=20)
    sphere.keep_immersed_part()
    sphere.mass = 300 * sphere.volume
    sphere.center_of_mass = np.array([0.0, 0.0, -0.2])
    sphere.add_all_rigid_body_dofs()
    K1 = sphere.compute_hydrostatic_stiffness()
    K2 = sphere.translated([1.0, 0.0, 0.0]).compute_hydrostatic_stiffness()
    assert np.allclose(K1, K2)

def test_non_neutrally_buoyant_inertia():
    body = cpt.VerticalCylinder(radius=1.0, length=1.0, center=(0.0, 0.0, -0.5), nx=20, ntheta=40, nr=20)
    body.rotation_center = (0, 0, 0)
    body.add_all_rigid_body_dofs()
    body.keep_immersed_part()
    body.center_of_mass = (0.0, 0.0, -0.25)
    body.mass = body.volume * 500
    M = body.compute_rigid_body_inertia().values
    assert np.allclose(np.diag(M), np.array([1570, 1570, 1570, 936, 936, 801]), rtol=5e-2)

def test_non_neutrally_buoyant_inertia_invariance_by_translation():
    sphere = cpt.Sphere(radius=1.0, center=(0, 0, 0), nphi=20, ntheta=20)
    sphere.rotation_center = np.array([0.0, 0.0, 0.0])
    sphere.mass = 300 * sphere.volume
    sphere.center_of_mass = np.array([0.0, 0.0, -0.2])
    sphere.add_all_rigid_body_dofs()
    M1 = sphere.compute_rigid_body_inertia()
    M2 = sphere.translated([1.0, -1.0, 10.0]).compute_rigid_body_inertia()
    print(M1.values)
    print(M2.values)
    assert np.allclose(M1, M2)



#! /usr/bin/env python3

import numpy as np
import capytaine as cpt
from time import time

radius = 2.
spacing = 5.
nb_cols = 14
nb_rows = 14

omega = 1.

x, y = np.meshgrid(np.arange(nb_cols)*spacing, np.arange(nb_rows)*spacing)
x = x.flatten()
y = y.flatten()

bodies = [cpt.FloatingBody(mesh=cpt.mesh_sphere(radius=radius,
                                                center=(x[ii], y[ii], 0),
                                                resolution=(10,10)))
            for ii in range(len(x))]

for body in bodies:
    body.keep_immersed_part()
    body.add_translation_dof(name='Heave')

# Dense problem setup and solution
print('Solving dense problem')
all_bodies = cpt.FloatingBody.join_bodies(*bodies, name="joined bodies")
problems = [cpt.RadiationProblem(body=all_bodies, radiating_dof=dof, omega=omega)
        for dof in all_bodies.dofs]
solver = cpt.BEMSolver()
dresults = solver.solve_all(problems)
ddata = cpt.assemble_dataset(dresults)


# Hierarchical problem setup
all_bodies = cpt.FloatingBody.cluster_bodies(*bodies, name="clustered bodies")
problems = [cpt.RadiationProblem(body=all_bodies, radiating_dof=dof, omega=omega)
            for dof in all_bodies.dofs]


cpt.set_logging(level='INFO')  # prints the number of iterations
# Hierarchical solve, with preconditioner
print('Solving hierarchical problem with preconditioner')
t0 = time()
sparse_engine = cpt.HierarchicalPrecondMatrixEngine(ACA_distance=2.0, ACA_tol=1e-2)
solver = cpt.BEMSolver(engine=sparse_engine)
presults = solver.solve_all(problems)
tP = time() - t0

data = cpt.assemble_dataset(presults)

# Hierarchical solve, no preconditioner
print('Solving hierarchical problem without preconditioner')
t0 = time()
sparse_engine = cpt.HierarchicalToeplitzMatrixEngine(ACA_distance=2.0, ACA_tol=1e-2)
solver = cpt.BEMSolver(engine=sparse_engine)
hresults = solver.solve_all(problems)
tNP = time() - t0

# The clustering process changes the order of the bodies with respect to
# the one in list 'bodies'. The correspondence with keys is maintained correct.
# However, if one wants to go back to the initial ordering, the following
# lines can be used.

ordered_dofs_names = []
for body in bodies:
    ordered_dofs_names += [body.name + '__' + dofkey for dofkey in list(body.dofs.keys())  ]
reordered_dofs = {'radiating_dof': ordered_dofs_names,
                  'influenced_dof': ordered_dofs_names}

reordered_data = data.reindex(reordered_dofs)

print('error = ', np.linalg.norm(presults[0].sources - hresults[0].sources))
print('time with preconditioner = ', tP)
print('time without preconditioner = ', tNP)

import matplotlib.pyplot as plt
fig, ax = plt.subplots(1, 3)
ax[0].matshow(ddata['added_mass'].data[0,:,:])
ax[1].matshow(reordered_data['added_mass'].data[0,:,:])
ax[2].matshow(data['added_mass'].data[0,:,:])
ax[0].set_title('Dense')
ax[1].set_title('HP, reordered')
ax[2].set_title('HP, before reordering')
fig.suptitle('Added mass matrix')

plt.show()



import pytest

import numpy as np
import capytaine as cpt

from capytaine.bem.solver import BEMSolver
from capytaine.green_functions.delhommeau import Delhommeau
from capytaine.bem.engines import BasicMatrixEngine
from capytaine.bem.problems_and_results import RadiationProblem
from capytaine.bodies.predefined.spheres import Sphere

sphere = Sphere(radius=1.0, ntheta=2, nphi=3, clip_free_surface=True)
sphere.add_translation_dof(direction=(1, 0, 0), name="Surge")
method = ['indirect','direct']
adjoint_double_layer = [True,False]

def test_cache_matrices():
    """Test how the BasicMatrixEngine caches the interaction matrices."""
    gf = Delhommeau()
    params_1 = (sphere.mesh, sphere.mesh, 0.0, np.inf, 1.0, gf)
    params_2 = (sphere.mesh, sphere.mesh, 0.0, np.inf, 2.0, gf)

    # No cache
    engine = BasicMatrixEngine(matrix_cache_size=0)
    S, K             = engine.build_matrices(*params_1)
    S_again, K_again = engine.build_matrices(*params_1)
    assert S is not S_again
    assert K is not K_again

    # Cache
    engine = BasicMatrixEngine(matrix_cache_size=1)
    S, K                     = engine.build_matrices(*params_1)
    S_again, K_again         = engine.build_matrices(*params_1)
    _, _                     = engine.build_matrices(*params_2)
    S_once_more, K_once_more = engine.build_matrices(*params_2)
    assert S is S_again
    assert S is not S_once_more
    assert K is K_again
    assert K is not K_once_more

@pytest.mark.parametrize("method", method)
def test_custom_linear_solver(method):
    """Solve a simple problem with a custom linear solver."""
    problem = RadiationProblem(body=sphere, omega=1.0, water_depth=np.inf)

    reference_solver = BEMSolver(
        engine=BasicMatrixEngine(linear_solver="gmres", matrix_cache_size=0)
    )
    reference_result = reference_solver.solve(problem,method=method)

    def my_linear_solver(A, b):
        """A dumb solver for testing."""
        return np.linalg.inv(A) @ b

    my_bem_solver = BEMSolver(
        engine=BasicMatrixEngine(linear_solver=my_linear_solver, matrix_cache_size=0)
    )
    assert 'my_linear_solver' in my_bem_solver.exportable_settings['linear_solver']

    result = my_bem_solver.solve(problem,method=method)
    assert np.isclose(reference_result.added_masses['Surge'], result.added_masses['Surge'])

@pytest.mark.parametrize("adjoint_double_layer", adjoint_double_layer)
def test_gradiant_G_shape(adjoint_double_layer):
    mesh = cpt.mesh_sphere(radius=1, center=(0, 0, 0), resolution=(10, 10)).immersed_part()
    S, gradG_1 = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False,adjoint_double_layer=adjoint_double_layer)
    assert gradG_1.shape == (mesh.nb_faces, mesh.nb_faces, 3)


def test_gradient_G_a_posteriori_scalar_product():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    S, gradG = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    S, K = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=True)

    # Scalar product with the faces normal vectors
    K_ = np.zeros_like(S)
    for i in range(mesh.nb_faces):
        for j in range(mesh.nb_faces):
            K_[i, j] = gradG[i, j, :] @ mesh.faces_normals[i, :]

    # Shorter implementiation of the exact same scalar products
    K__ = np.einsum('...jk,...k->...j', gradG, mesh.faces_normals)

    np.testing.assert_allclose(K, K_)
    np.testing.assert_allclose(K, K__)


def test_gradient_G_a_posteriori_scalar_product_directBIE():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    S, gradG = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False, adjoint_double_layer=False)
    S, K = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=True, adjoint_double_layer=False)

    # Scalar product with the faces normal vectors
    K_ = np.zeros_like(S)
    for i in range(mesh.nb_faces):
        for j in range(mesh.nb_faces):
            K_[i, j] = gradG[i, j, :] @ mesh.faces_normals[j, :]

    # Shorter implementiation of the exact same scalar products
    K__ = np.einsum('...k,...k->...', gradG, mesh.faces_normals)

    np.testing.assert_allclose(K, K_)
    np.testing.assert_allclose(K, K__)
    np.testing.assert_allclose(K_, K__)


def test_gradient_G_with_collocation_points():
    mesh = cpt.mesh_sphere(radius=1, center=(0, 0, 0), resolution=(10, 10)).immersed_part()
    _, gradG_1 = cpt.Delhommeau().evaluate(mesh.faces_centers, mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    _, gradG_2 = cpt.Delhommeau().evaluate(mesh.copy(), mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    np.testing.assert_allclose(gradG_1, gradG_2, atol=1e-10, rtol=1e-10)


@pytest.mark.parametrize("adjoint_double_layer", adjoint_double_layer)
def test_gradient_G_diagonal_term(adjoint_double_layer):
    mesh = cpt.mesh_sphere(radius=1, center=(0, 0, 0), resolution=(10, 10)).immersed_part()
    # Passing two different Python objects
    _, gradG_1 = cpt.Delhommeau().evaluate(mesh.copy(), mesh, 0.0, np.inf, 1.0,
					early_dot_product=False, adjoint_double_layer=adjoint_double_layer)
    # Passing the same Python object
    _, gradG_2 = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0,
					early_dot_product=False, adjoint_double_layer=adjoint_double_layer)

    diag_normal = np.zeros_like(gradG_2)
    for i in range(mesh.nb_faces):
        diag_normal[i, i, :] = mesh.faces_normals[i, :]

    np.testing.assert_allclose(gradG_1, gradG_2 - 0.5*diag_normal, atol=1e-10, rtol=1e-10)


def test_a_posteriori_scalar_product_direct_method():
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    S, gradG = cpt.Delhommeau().evaluate(mesh, mesh, 0.0, np.inf, 1.0, early_dot_product=False)
    D_ = np.zeros_like(S)
    for i in range(mesh.nb_faces):
        for j in range(mesh.nb_faces):
            D_[i, j] = gradG[i, j, :] @ mesh.faces_normals[j, :]
    D__ = np.einsum('...k,...k->...', gradG, mesh.faces_normals)
    np.testing.assert_allclose(D_, D__)



"""This modules implements block Toeplitz matrices to be used in Hierarchical Toeplitz matrices.

The module also contains several special cases such as block symmetric Toeplitz matrices and block circulant matrices.
All classes inherits from the BlockMatrix class.
"""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging
from typing import Tuple, List, Set, Iterable

import numpy as np

from capytaine.matrices.block import BlockMatrix

LOG = logging.getLogger(__name__)


################################################################################
#                            Block Toeplitz matrix                             #
################################################################################

class BlockToeplitzMatrix(BlockMatrix):
    """A (2D) block Toeplitz matrix, stored as a list of blocks.
    All blocks should have the same shape.

    Stored in the backend as a 1×(2N-1) array of arrays."""

    # INITIALIZATION

    def _compute_shape(self) -> Tuple[int, int]:
        # The full shape is found by multiplying the shape of the blocks. All of them have the same shape.
        return (self._stored_block_shapes[0][0]*self.nb_blocks[0],
                self._stored_block_shapes[1][0]*self.nb_blocks[1])

    def _compute_nb_blocks(self) -> Tuple[int, int]:
        """Will be overridden by subclasses."""
        assert self._stored_nb_blocks[1] % 2 == 1, "Expecting an odd number of blocks to build a Toeplitz matrix"
        n = (self._stored_nb_blocks[1]+1)//2
        return n, n

    def _check_dimensions_of_blocks(self) -> bool:
        for block in self._stored_blocks[0, :]:
            if not block.shape == self.block_shape:  # All blocks have same shape
                return False
        return True

    # ACCESSING DATA

    @property
    def block_shapes(self) -> Tuple[List[int], List[int]]:
        """The shapes of the blocks composing the block matrix.
        Actually, they should be all the same."""
        return ([self._stored_block_shapes[0][0]]*self.nb_blocks[0],
                [self._stored_block_shapes[1][0]]*self.nb_blocks[1])

    @property
    def block_shape(self) -> Tuple[int, int]:
        """The shape of any of the blocks."""
        return self._stored_block_shapes[0][0], self._stored_block_shapes[1][0]

    def _block_indices_of(self, k: int) -> Set[Tuple[int, int]]:
        """The block indices at which the stored block k can be found in the full matrix of size n.
        Will be overridden by subclasses."""
        n = self.nb_blocks[0]

        if k < n:
            i, j = 0, k  # Upper triangle
        elif n <= k < 2*n:
            i, j = 2*n-1-k, 0  # Lower triangle
        else:
            raise AttributeError

        indices = set()
        while i < n and j < n:
            indices.add((i, j))
            i, j = i+1, j+1  # Going along the diagonal

        return indices

    @property
    def all_blocks(self):
        all_blocks = np.empty(self.nb_blocks, dtype=object)
        for k in range(self._stored_nb_blocks[1]):
            for i, j in self._block_indices_of(k):
                all_blocks[i, j] = self._stored_blocks[0, k]
        return all_blocks

    def _positions_of(self, k: int, global_frame=(0, 0)) -> List[Tuple[int, int]]:
        """The positions in the full matrix at which the block k from the first line can also be found."""
        shape = self.block_shape
        return sorted([(global_frame[0] + i*shape[0], global_frame[1] + j*shape[1])
                       for i, j in self._block_indices_of(k)])

    def _stored_block_positions(self, global_frame=(0, 0)) -> Iterable[List[Tuple[int, int]]]:
        """The position of each blocks in the matrix.

        Example::

            AABB
            AABB  ->  list(matrix._stored_block_positions) = [[(0,0), (2, 2)], [(0, 2), (2, 0)]]
            BBAA
            BBAA
        """
        return (self._positions_of(k, global_frame=global_frame) for k in range(self._stored_nb_blocks[1]))

    # # TRANSFORMING DATA

    @property
    def circulant_super_matrix(self):
        if not hasattr(self, '_circulant_super_matrix'):
            self._circulant_super_matrix = BlockCirculantMatrix(
                self._stored_blocks,
                _stored_block_shapes=self._stored_block_shapes,
                check=False)
        return self._circulant_super_matrix

    def matvec(self, other):
        """Matrix vector product.
        Named as such to be used as scipy LinearOperator."""
        LOG.debug(f"Product of {self} with vector of shape {other.shape}")
        A = self.circulant_super_matrix
        b = np.concatenate([other, np.zeros(A.shape[1] - self.shape[1])])
        return (A @ b)[:self.shape[0]]

    def rmatvec(self, other):
        """Matrix vector product.
        Named as such to be used as scipy LinearOperator."""
        LOG.debug(f"Product of vector of shape {other.shape} with {self}")
        if other.ndim == 2 and other.shape[0] == 1:  # Actually a 1×N matrix
            other = other[0, :]
        A = self.circulant_super_matrix
        b = np.concatenate([other, np.zeros(A.shape[0] - self.shape[0])])
        return (A.rmatvec(b))[:self.shape[1]]


################################################################################
#                       Block symmetric Toeplitz matrix                        #
################################################################################

class BlockSymmetricToeplitzMatrix(BlockToeplitzMatrix):
    """A (2D) block symmetric Toeplitz matrix, stored as a list of blocks.
    All blocks should have the same shape.

    Stored in the backend as a 1×N array of arrays."""

    def _compute_nb_blocks(self) -> Tuple[int, int]:
        n = self._stored_nb_blocks[1]
        return n, n

    def _block_indices_of(self, k: int) -> Set[Tuple[int, int]]:
        """The block indices at which the stored block k can be found in the full matrix."""
        n = self.nb_blocks[0]
        assert k < n

        if k == 0:
            return BlockToeplitzMatrix._block_indices_of(self, 0)
        else:
            return BlockToeplitzMatrix._block_indices_of(self, k).union(
                BlockToeplitzMatrix._block_indices_of(self, 2*n-k-1))

    @property
    def circulant_super_matrix(self):
        if not hasattr(self, '_circulant_super_matrix'):
            self._circulant_super_matrix = EvenBlockSymmetricCirculantMatrix(
                self._stored_blocks,
                _stored_block_shapes=self._stored_block_shapes,
                check=False)
        return self._circulant_super_matrix


################################################################################
#                            Block circulant matrix                            #
################################################################################

class BlockCirculantMatrix(BlockToeplitzMatrix):
    """A (2D) block circulant matrix, stored as a list of blocks.
    All blocks should have the same shape.

    Stored in the backend as a 1×N array of arrays."""

    def _compute_nb_blocks(self) -> Tuple[int, int]:
        n = self._stored_nb_blocks[1]
        return n, n

    def _block_indices_of(self, k: int) -> Set[Tuple[int, int]]:
        """The block indices at which the stored block k can be found in the full matrix."""
        n = self.nb_blocks[0]
        assert k < n

        if k == 0:
            return BlockToeplitzMatrix._block_indices_of(self, 0)
        else:
            return BlockToeplitzMatrix._block_indices_of(self, k).union(
                BlockToeplitzMatrix._block_indices_of(self, n+k-1))

    # LINEAR SYSTEMS

    def block_diagonalize(self):
        """Returns a vector of matrices"""
        if not hasattr(self, 'block_diagonalization'):
            if all(isinstance(matrix, BlockMatrix) for matrix in self._stored_blocks[0, :]):
                self.block_diagonalization = BlockMatrix.fft_of_list(*self.all_blocks[:, 0])
            else:
                stacked_blocks = np.empty((self.nb_blocks[1],) + self.block_shape, dtype=self.dtype)
                for i, block in enumerate(self.all_blocks[:, 0]):
                    stacked_blocks[i] = block.full_matrix() if not isinstance(block, np.ndarray) else block
                self.block_diagonalization =  np.fft.fft(stacked_blocks, axis=0)
        return self.block_diagonalization

    def matvec(self, other):
        """Matrix vector product.
        Named as such to be used as scipy LinearOperator."""
        LOG.debug(f"Product of {self} with vector of shape {other.shape}")
        fft_of_vector = np.fft.fft(np.reshape(other, (self.nb_blocks[0], self.block_shape[1], 1)), axis=0)
        blocks_of_diagonalization = self.block_diagonalize()
        try:  # Try to run it as vectorized numpy arrays.
            fft_of_result = blocks_of_diagonalization @ fft_of_vector
        # When the above fails, numpy 1.15 returns a TypeError, whereas numpy 1.16 returns a ValueError.
        except (TypeError, ValueError):  # Or do the same thing with list comprehension.
            fft_of_result = np.array([block @ vec for block, vec in zip(blocks_of_diagonalization, fft_of_vector)])
        result = np.fft.ifft(fft_of_result, axis=0).reshape(self.shape[0])
        if self.dtype == complex or other.dtype == complex:
            return np.asarray(result)
        else:
            return np.asarray(np.real(result))

    def rmatvec(self, other):
        """Matrix vector product.
        Named as such to be used as scipy LinearOperator."""
        other = np.conjugate(other)
        fft_of_vector = np.fft.ifft(np.reshape(other, (self.nb_blocks[0], 1, self.block_shape[0])), axis=0)
        blocks_of_diagonalization = self.block_diagonalize()
        try:  # Try to run it as vectorized numpy arrays.
            fft_of_result = fft_of_vector @ blocks_of_diagonalization
        # When the above fails, numpy 1.15 returns a TypeError, whereas numpy 1.16 returns a ValueError.
        except (TypeError, ValueError):
            # Instead we do the same thing with list comprehension.
            fft_of_result = np.array(
                [block.rmatvec(vec.flatten()) for block, vec in zip(blocks_of_diagonalization, fft_of_vector)]
            )
        result = np.fft.fft(fft_of_result, axis=0).reshape(self.shape[1])
        if self.dtype == complex or other.dtype == complex:
            return np.asarray(result)
        else:
            return np.asarray(np.real(result))


###########################################################################
#                    Block symmetric circulant matrix                     #
###########################################################################

class EvenBlockSymmetricCirculantMatrix(BlockCirculantMatrix, BlockSymmetricToeplitzMatrix):
    """A block symmetric circulant matrix, with an even number of blocks.

    Examples::

        ABCB
        BABC
        CBAB
        BCBA

        ABCDCB
        BABCDB
        CBABCD
        DCBABC
        CDCBAB
        BCDCBA

    Stored in the backend as a 1×(N/2+1) array of arrays."""

    def _compute_nb_blocks(self) -> Tuple[int, int]:
        """The number of blocks in the full matrix."""
        n = (self._stored_nb_blocks[1] - 1)*2
        return n, n

    def _block_indices_of(self, k: int) -> List[Tuple[int, int]]:
        n = self.nb_blocks[0]
        assert k < n/2 + 1
        if k == 0:
            return BlockToeplitzMatrix._block_indices_of(self, 0)
        else:
            return (BlockToeplitzMatrix._block_indices_of(self, k) |
                BlockToeplitzMatrix._block_indices_of(self, n+k-1) |
                BlockToeplitzMatrix._block_indices_of(self, n-k)   |
                BlockToeplitzMatrix._block_indices_of(self, 2*n-k-1)
                    )


class OddBlockSymmetricCirculantMatrix(BlockCirculantMatrix, BlockSymmetricToeplitzMatrix):
    """A block symmetric circulant matrix, with an odd number of blocks.

    Examples::

        ABCCB
        BABCC
        CBABC
        CCBAB
        BCCBA

        ABCDDCB
        BABCDDB
        CBABCDD
        DCBABCD
        DDCBABC
        CDDCBAB
        BCDDCBA

    Stored in the backend as a 1×(N+1)/2 array of arrays."""

    def _compute_nb_blocks(self) -> Tuple[int, int]:
        n = self._stored_nb_blocks[1]*2 - 1
        return n, n

    def _block_indices_of(self, k: int) -> List[Tuple[int, int]]:
        n = self.nb_blocks[0]
        assert k < (n+1)/2
        if k == 0:
            return BlockToeplitzMatrix._block_indices_of(self, 0)
        else:
            return (BlockToeplitzMatrix._block_indices_of(self, k) |
                BlockToeplitzMatrix._block_indices_of(self, n+k-1) |
                BlockToeplitzMatrix._block_indices_of(self, n-k)   |
                BlockToeplitzMatrix._block_indices_of(self, 2*n-k-1)
                    )



#!/usr/bin/env python
# coding: utf-8
"""Tests related to post-processing computation of potential, velocity and free surface elevation."""

import pytest
from pytest import approx
import numpy as np
import capytaine as cpt

#######################################################################
#                  Test shapes of inputs and outputs                  #
#######################################################################

@pytest.fixture
def solver():
    return cpt.BEMSolver()

@pytest.fixture
def result(solver):
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(name="Heave")
    result = solver.solve(cpt.DiffractionProblem(body=body, omega=1.0, wave_direction=np.pi/4), keep_details=True)
    return result


##################
#  Single point  #
##################

def test_airy_waves_potential_at_point(result):
    from capytaine.bem.airy_waves import airy_waves_potential
    point = (0.0, 0.0, -3.0)
    phi = airy_waves_potential(point, result)
    assert phi.shape == (1,)

def test_compute_potential_at_point(solver, result):
    point = (0.0, 0.0, -3.0)
    phi = solver.compute_potential(point, result)
    assert phi.shape == (1,)

def test_airy_waves_velocity_at_point(result):
    from capytaine.bem.airy_waves import airy_waves_velocity
    point = (0.0, 0.0, -3.0)
    u = airy_waves_velocity(point, result)
    assert u.shape == (1, 3)

def test_compute_velocity_at_point(solver, result):
    point = (0.0, 0.0, -3.0)
    u = solver.compute_velocity(point, result)
    assert u.shape == (1, 3)

def test_airy_waves_free_surface_elevation_at_point(result):
    from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
    point = (0.0, 3.0)
    fse = airy_waves_free_surface_elevation(point, result)
    assert fse.shape == (1,)
    point_ = (0.0, 3.0, 0.0)
    fse_ = airy_waves_free_surface_elevation(point_, result)
    assert fse_.shape == (1,)
    assert np.allclose(fse, fse_)

def test_compute_free_surface_elevation_at_point(solver, result):
    point = (0.0, 3.0)
    fse = solver.compute_free_surface_elevation(point, result)
    assert fse.shape == (1,)
    point_ = (0.0, 3.0, 0.0)
    fse_ = solver.compute_free_surface_elevation(point_, result)
    assert fse_.shape == (1,)
    assert np.allclose(fse, fse_)


#################
#  Points list  #
#################

def test_airy_waves_potential_at_points_list(result):
    from capytaine.bem.airy_waves import airy_waves_potential
    points = [(0.0, 0.0, -3.0), (0.0, 1.0, -2.0), (1.0, 1.0, -1.0)]
    phi = airy_waves_potential(points, result)
    assert phi.shape == (3,)

def test_compute_potential_at_points_list(solver, result):
    points = [(0.0, 0.0, -3.0), (0.0, 1.0, -2.0), (1.0, 1.0, -1.0)]
    phi = solver.compute_potential(points, result)
    assert phi.shape == (3,)

def test_airy_waves_velocity_at_points_list(result):
    from capytaine.bem.airy_waves import airy_waves_velocity
    points = [(0.0, 0.0, -3.0), (0.0, 1.0, -2.0), (1.0, 1.0, -1.0)]
    u = airy_waves_velocity(points, result)
    assert u.shape == (3, 3)

def test_compute_velocity_at_points_list(solver, result):
    points = [(0.0, 0.0, -3.0), (0.0, 1.0, -2.0), (1.0, 1.0, -1.0)]
    u = solver.compute_velocity(points, result)
    assert u.shape == (3, 3)

def test_airy_waves_free_surface_elevation_at_points_list(result):
    from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
    points = [(0.0, 3.0), (1.0, -2.0), (1.0, -1.0)]
    fse = airy_waves_free_surface_elevation(points, result)
    assert fse.shape == (3,)
    points_ = [(0.0, 3.0, 0.0), (1.0, -2.0, 0.0), (1.0, -1.0, 0.0)]
    fse_ = airy_waves_free_surface_elevation(points_, result)
    assert fse_.shape == (3,)
    assert np.allclose(fse, fse_)

def test_compute_free_surface_elevation_at_points_list(solver, result):
    points = [(0.0, 3.0), (1.0, -2.0), (1.0, -1.0)]
    fse = solver.compute_free_surface_elevation(points, result)
    assert fse.shape == (3,)
    points_ = [(0.0, 3.0, 0.0), (1.0, -2.0, 0.0), (1.0, -1.0, 0.0)]
    fse_ = solver.compute_free_surface_elevation(points_, result)
    assert fse_.shape == (3,)
    assert np.allclose(fse, fse_)


##############
#  Meshgrid  #
##############

def test_airy_waves_potential_at_meshgrid(result):
    from capytaine.bem.airy_waves import airy_waves_potential
    points = np.meshgrid(np.linspace(0.0, 1.0, 2), np.linspace(-1.0, 1.0, 3), np.linspace(-2.0, 0.0, 4))
    phi = airy_waves_potential(points, result)
    assert phi.shape == points[0].shape

def test_compute_potential_at_meshgrid(solver, result):
    points = np.meshgrid(np.linspace(2.0, 3.0, 2), np.linspace(-1.0, 1.0, 3), np.linspace(-2.0, -1.0, 4))
    phi = solver.compute_potential(points, result)
    assert phi.shape == points[0].shape

def test_airy_waves_velocity_at_meshgrid(result):
    from capytaine.bem.airy_waves import airy_waves_velocity
    points = np.meshgrid(np.linspace(0.0, 1.0, 2), np.linspace(-1.0, 1.0, 3), np.linspace(-2.0, 0.0, 4))
    u = airy_waves_velocity(points, result)
    assert u.shape == (*points[0].shape, 3)

def test_compute_velocity_at_meshgrid(solver, result):
    points = np.meshgrid(np.linspace(2.0, 3.0, 2), np.linspace(4.0, 5.0, 3), np.linspace(-2.0, -1.0, 4))
    u = solver.compute_velocity(points, result)
    assert u.shape == (*points[0].shape, 3)

def test_airy_waves_free_surface_elevation_at_meshgrid(result):
    from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
    points = np.meshgrid(np.linspace(0.0, 1.0, 2), np.linspace(-1.0, 1.0, 3))
    fse = airy_waves_free_surface_elevation(points, result)
    assert fse.shape == points[0].shape
    points_ = np.meshgrid(np.linspace(0.0, 1.0, 2), np.linspace(-1.0, 1.0, 3), np.linspace(0.0, 0.0, 1))
    fse_ = airy_waves_free_surface_elevation(points_, result)
    assert fse_.shape == points_[0].shape
    assert np.allclose(fse, fse_.squeeze())

def test_compute_free_surface_elevation_at_meshgrid(solver, result):
    points = np.meshgrid(np.linspace(2.0, 3.0, 2), np.linspace(3.0, 4.0, 3))
    fse = solver.compute_free_surface_elevation(points, result)
    assert fse.shape == points[0].shape
    points_ = np.meshgrid(np.linspace(2.0, 3.0, 2), np.linspace(3.0, 4.0, 3), np.linspace(0.0, 0.0, 1))
    fse_ = solver.compute_free_surface_elevation(points_, result)
    assert fse_.shape == points_[0].shape
    assert np.allclose(fse, fse_.squeeze())


##########
#  Mesh  #
##########

def test_airy_waves_potential_on_mesh(result):
    from capytaine.bem.airy_waves import airy_waves_potential
    mesh = cpt.mesh_vertical_cylinder(radius=2.0, length=5.0).immersed_part()
    phi = airy_waves_potential(mesh, result)
    assert phi.shape == (mesh.nb_faces,)

def test_compute_potential_on_mesh(solver, result):
    mesh = cpt.mesh_vertical_cylinder(radius=2.0, length=5.0).immersed_part()
    phi = solver.compute_potential(mesh, result)
    assert phi.shape == (mesh.nb_faces,)

def test_airy_waves_velocity_on_mesh(result):
    from capytaine.bem.airy_waves import airy_waves_velocity
    mesh = cpt.mesh_vertical_cylinder(radius=2.0, length=5.0).immersed_part()
    u = airy_waves_velocity(mesh, result)
    assert u.shape == (mesh.nb_faces, 3)

def test_compute_velocity_on_mesh(solver, result):
    mesh = cpt.mesh_vertical_cylinder(radius=2.0, length=5.0).immersed_part()
    u = solver.compute_velocity(mesh, result)
    assert u.shape == (mesh.nb_faces, 3)

def test_airy_waves_free_surface_elevation_on_mesh(result):
    from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
    mesh = cpt.mesh_rectangle(center=(0, 0, 0), normal=(0, 0, -1), resolution=(3, 3), size=(2.0, 2.0))
    fse = airy_waves_free_surface_elevation(mesh, result)
    assert fse.shape == (mesh.nb_faces,)

def test_compute_free_surface_elevation_on_mesh(solver, result):
    mesh = cpt.mesh_rectangle(center=(0, 0, 0), normal=(0, 0, -1), resolution=(3, 3), size=(2.0, 2.0))
    fse = solver.compute_free_surface_elevation(mesh, result)
    assert fse.shape == (mesh.nb_faces,)


#################
#  FreeSurface  #
#################

def test_airy_waves_free_surface_elevation_on_free_surface(result):
    from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
    fs = cpt.FreeSurface(nx=3, ny=3)
    fse = airy_waves_free_surface_elevation(fs, result)
    assert fse.shape == (fs.mesh.nb_faces,)

def test_compute_free_surface_elevation_on_free_surface(solver, result):
    fs = cpt.FreeSurface(nx=3, ny=3)
    fse = solver.compute_free_surface_elevation(fs, result)
    assert fse.shape == (fs.mesh.nb_faces,)


#######################################################################
#                            Check values                             #
#######################################################################

def test_pressure_integration(solver, result):
    f = result.body.integrate_pressure(solver.compute_pressure(result.body.mesh, result))
    assert f == pytest.approx(result.forces)

def test_reconstruction_of_given_boundary_condition(solver, result):
    velocities = solver.compute_velocity(result.body.mesh, result)
    normal_velocities = np.einsum('...k,...k->...', velocities, result.body.mesh.faces_normals)
    np.testing.assert_allclose(normal_velocities, result.problem.boundary_condition)

def test_airy_wave_free_surface_elevation_values():
    from capytaine.bem.airy_waves import airy_waves_free_surface_elevation
    pb = cpt.DiffractionProblem(wave_direction=0.0, wavelength=1.0)

    assert np.isclose(np.real(airy_waves_free_surface_elevation([0, 0], pb)), 1.0)
    assert np.isclose(np.real(airy_waves_free_surface_elevation([0.25, 0], pb)), 0.0, atol=1e-5)
    assert np.isclose(np.real(airy_waves_free_surface_elevation([0.5, 0], pb)), -1.0)

def test_integrated_pressure(solver, result):
    pressure = solver.compute_pressure(result.body.mesh, result)
    forces = result.body.integrate_pressure(pressure)
    assert result.forces == approx(forces)

def test_fse_zero_frequency(solver):
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(name="Heave")
    pb = cpt.RadiationProblem(body=body, omega=0.0, radiating_dof="Heave")
    res = solver.solve(pb, keep_details=True)
    points = -np.random.rand(10, 3)
    pot = solver.compute_potential(points, res)
    fse = solver.compute_free_surface_elevation(points, res)
    with pytest.raises(TypeError):
        vel = solver.compute_velocity(points, res)

def test_direct_solver(solver):
    mesh = cpt.mesh_sphere(resolution=(4, 4)).immersed_part()
    body = cpt.FloatingBody(mesh=mesh)
    body.add_translation_dof(name="Heave")
    res = solver.solve(cpt.DiffractionProblem(body=body, omega=1.0, wave_direction=np.pi/4), keep_details=True, method='direct')
    points = [(0.0, 0.0, -3.0), (0.0, 1.0, -2.0), (1.0, 1.0, -1.0)]
    with pytest.raises(Exception, match="direct method"):
        pot = solver.compute_potential(points, res)




#!/usr/bin/env python
# coding: utf-8
"""Test of matplotlib user interface methods."""

import pytest
import numpy as np
import capytaine as cpt

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

@pytest.mark.skipif(plt is None,
                     reason='matplotlib is not installed')
def test_showmatplotlib_with_colors(tmp_path):
    cylinder = cpt.HorizontalCylinder(
        length=5.0, radius=1.0, center=(0, 0, -2), nr=5, nx=20, ntheta=20,
    )

    problem = cpt.DiffractionProblem(body=cylinder, wave_direction=0.0, omega=1.0)
    solver = cpt.BEMSolver()
    results = solver.solve(problem)

    cylinder.show_matplotlib(saveas=tmp_path / "showmpl.png")

    fig = plt.figure()
    ax = fig.add_subplot(111,projection='3d')
    from matplotlib import colormaps
    colormap = colormaps['cividis']
    cylinder.show_matplotlib(color_field=np.abs(results.potential),
                             ax=ax, cbar_label=r'$\phi (m^2/s)$')
    ax.set_title('Potential distribution')
    plt.savefig(tmp_path / "showmpl_with_color_field.png")



import logging

import numpy as np

LOG = logging.getLogger(__name__)

def _get_water_depth(free_surface, water_depth, sea_bottom, default_water_depth=np.inf):
    if water_depth is None and sea_bottom is None:
        return default_water_depth
    elif water_depth is not None and sea_bottom is None:
        if water_depth <= 0.0:
            raise ValueError(f"`water_depth` should be strictly positive. Received value: {water_depth}")
        return float(water_depth)
    elif water_depth is None and sea_bottom is not None:
        LOG.warning("To uniformize notations througouth Capytaine, setting `water_depth` is preferred to `sea_bottom` since version 2.0.")
        return float(free_surface - sea_bottom)
    else:
        raise ValueError("Cannot give both a `water_depth` and a `sea_bottom`.")



import os
import tempfile
import nox

# Create virtual environments in a temporary directory somewhere else because
# meson-python does not like local virtualenvironments
# (https://github.com/capytaine/capytaine/issues/396)
nox.options.envdir = os.path.join(tempfile.gettempdir(), "nox-capytaine")

os.environ.update({"PDM_IGNORE_SAVED_PYTHON": "1"})
os.environ.update({"PDM_USE_VENV": "1"})

NOXFILE_DIR = os.path.dirname(__file__)

EXAMPLE_FILES = ["compare_Green_functions.py", "convergence_study.py", "custom_dofs.py", "custom_Green_function.py", "finite_depth_cylinder.py", "free_surface_elevation.py", "haskind.py", "kochin.py", "multibody.py", "plot_influence_matrix.py", "plot_velocity_in_domain.py", "radiation_cylinder.py"]

def run_tests(session):
    with session.chdir(session.create_tmp()):
        session.run("python", "-m", "pytest", os.path.join(NOXFILE_DIR, "pytest"))
        session.run('python', '-c', '"import capytaine; print(capytaine.__version__)"')
        session.run('capytaine', '--help')
        session.run('capytaine', os.path.join(NOXFILE_DIR, "pytest", "Nemoh_verification_cases", "Cylinder", "Nemoh.cal"))
        session.run('capytaine', os.path.join(NOXFILE_DIR, "pytest", "Nemoh_verification_cases", "Cylinder", "Nemoh_v3.cal"))
        for example_file in EXAMPLE_FILES:
            session.run('python', os.path.join(NOXFILE_DIR, "examples", example_file), env={'MPLBACKEND': 'pdf'})


@nox.session
def build_and_test_on_latest_env(session):
    session.install(".[test,optional]")
    run_tests(session)


@nox.session
def build_and_test_on_nightly_builds(session):
    # https://scientific-python.org/specs/spec-0004/
    session.install("--pre", "--upgrade",
                    "--extra-index-url", "https://pypi.anaconda.org/scientific-python-nightly-wheels/simple",
                    "numpy", "scipy", "pandas", "xarray")
    session.install("meson-python", "ninja", "charset-normalizer")
    session.install("--no-build-isolation", "-e", ".[test,optional]")
    run_tests(session)


@nox.session
@nox.parametrize("env_file", ["2023-11-07"])
def build_and_test_on_locked_env(session, env_file):
    session.install("pdm")
    session.run_always('pdm', 'sync', '--no-self', '-L', f"pytest/envs/{env_file}.lock")
    # Lock file was created with
    # pdm lock -d -G build -G test -G optional -L pytest/envs/2023-11-07.lock
    session.install("--no-build-isolation", "--no-deps", "-e", ".")
    # Editable install using pip and not pdm because pdm's editable install
    # does not seems to work with meson-python.
    run_tests(session)



"""Experimental function to compute the Response Amplitude Operator."""
# Copyright (C) 2017-2019 Matthieu Ancellin
# See LICENSE file at <https://github.com/mancellin/capytaine>

import logging

import numpy as np
import xarray as xr
from capytaine.post_pro.impedance import rao_transfer_function

LOG = logging.getLogger(__name__)


def rao(dataset, wave_direction=None, dissipation=None, stiffness=None):
    """Response Amplitude Operator.

    Parameters
    ----------
    dataset: xarray Dataset
        The hydrodynamical dataset.
        This function supposes that variables named 'inertia_matrix' and 'hydrostatic_stiffness' are in the dataset.
        Other variables can be computed by Capytaine, by those two should be manually added to the dataset.
    wave_direction: float, optional
        Select a wave directions for the computation. (Not recommended, kept for legacy.)
        Default: all wave directions in the dataset.
    dissipation: array, optional
        An optional dissipation matrix (e.g. Power Take Off) to be included in the RAO.
        Default: none.
    stiffness: array, optional
        An optional stiffness matrix (e.g. mooring stiffness) to be included in the RAO.
        Default: none.

    Returns
    -------
    xarray DataArray
        The RAO as an array depending of omega and the degree of freedom.
    """

    # ASSEMBLE MATRICES
    H = rao_transfer_function(dataset, dissipation, stiffness)
    fex = dataset.excitation_force

    LOG.info("Compute RAO.")

    # SOLVE LINEAR SYSTEMS
    # Match dimensions of the arrays to be sure to solve the right systems.
    H, fex = xr.broadcast(H, fex, exclude=["radiating_dof", "influenced_dof"])
    H = H.transpose(..., 'radiating_dof', 'influenced_dof')
    fex = fex.transpose(...,  'influenced_dof')

    if wave_direction is not None:  # Legacy behavior for backward compatibility
        H = H.sel(wave_direction=wave_direction)
        fex = fex.sel(wave_direction=wave_direction)

    # Solve and add coordinates
    rao_dims = [d for d in H.dims if d != 'influenced_dof']
    rao_coords = {c: H.coords[c] for c in H.coords if c != 'influenced_dof'}
    rao = xr.DataArray(np.linalg.solve(H, fex), coords=rao_coords, dims=rao_dims)

    return rao



#!/usr/bin/env python
# coding: utf-8
"""Tests for the resolution of the BEM problems using advanced techniques."""

import pytest

import numpy as np

import capytaine as cpt
from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.meshes.symmetric import AxialSymmetricMesh, ReflectionSymmetricMesh, TranslationalSymmetricMesh

from capytaine.bodies import FloatingBody
from capytaine.bodies.predefined.spheres import Sphere
from capytaine.bodies.predefined.cylinders import HorizontalCylinder, VerticalCylinder

from capytaine.bem.problems_and_results import RadiationProblem
from capytaine.io.xarray import assemble_dataset

from capytaine.meshes.geometry import xOz_Plane, yOz_Plane

from capytaine.matrices.low_rank import LowRankMatrix

solver_with_sym = cpt.BEMSolver(engine=cpt.HierarchicalToeplitzMatrixEngine(ACA_distance=8, matrix_cache_size=0))
solver_without_sym = cpt.BEMSolver(engine=cpt.BasicMatrixEngine(matrix_cache_size=0))
method = ['indirect','direct']
adjoint_double_layer = [True,False]

# Use a single solver in the whole module to avoid reinitialisation of the solver (0.5 second).
# Do not use a matrix cache in order not to risk influencing a test with another.


@pytest.mark.parametrize("method", method)
@pytest.mark.parametrize("depth", [10.0, np.inf])
@pytest.mark.parametrize("omega", [0.1, 10.0])
def test_floating_sphere(depth, omega, method):
    """Comparison of the added mass and radiation damping
    for a heaving sphere described using several symmetries
    in finite and infinite depth.
    """
    reso = 2

    full_sphere = Sphere(radius=1.0, ntheta=reso, nphi=4*reso, axial_symmetry=False, clip_free_surface=True)
    full_sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=full_sphere, omega=omega, water_depth=depth)
    result1 = solver_with_sym.solve(problem,method=method)

    half_sphere_mesh = full_sphere.mesh.extract_faces(
        np.where(full_sphere.mesh.faces_centers[:, 1] > 0)[0],
        name="half_sphere_mesh")
    two_halves_sphere = FloatingBody(ReflectionSymmetricMesh(half_sphere_mesh, xOz_Plane))
    two_halves_sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=two_halves_sphere, omega=omega, water_depth=depth)
    result2 = solver_with_sym.solve(problem,method=method)

    quarter_sphere_mesh = half_sphere_mesh.extract_faces(
        np.where(half_sphere_mesh.faces_centers[:, 0] > 0)[0],
        name="quarter_sphere_mesh")
    four_quarter_sphere = FloatingBody(ReflectionSymmetricMesh(ReflectionSymmetricMesh(quarter_sphere_mesh, yOz_Plane), xOz_Plane))
    assert 'None' not in four_quarter_sphere.mesh.tree_view()
    four_quarter_sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=four_quarter_sphere, omega=omega, water_depth=depth)
    result3 = solver_with_sym.solve(problem,method=method)

    clever_sphere = Sphere(radius=1.0, ntheta=reso, nphi=4*reso, axial_symmetry=True, clip_free_surface=True)
    clever_sphere.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=clever_sphere, omega=omega, water_depth=depth)
    result4 = solver_with_sym.solve(problem,method=method)

    # (quarter_sphere + half_sphere + full_sphere + clever_sphere).show()

    volume = 4/3*np.pi
    assert np.isclose(result1.added_masses["Heave"], result2.added_masses["Heave"], atol=1e-4*volume*problem.rho)
    assert np.isclose(result1.added_masses["Heave"], result3.added_masses["Heave"], atol=1e-4*volume*problem.rho)
    assert np.isclose(result1.added_masses["Heave"], result4.added_masses["Heave"], atol=1e-4*volume*problem.rho)
    assert np.isclose(result1.radiation_dampings["Heave"], result2.radiation_dampings["Heave"], atol=1e-4*volume*problem.rho)
    assert np.isclose(result1.radiation_dampings["Heave"], result3.radiation_dampings["Heave"], atol=1e-4*volume*problem.rho)
    assert np.isclose(result1.radiation_dampings["Heave"], result4.radiation_dampings["Heave"], atol=1e-4*volume*problem.rho)


def test_two_vertical_cylinders():
    distance = 5

    buoy = VerticalCylinder(length=3, radius=0.5, center=(-distance/2, -1, 0), nx=8, nr=3, ntheta=8)
    buoy.mesh = buoy.mesh.merged()
    buoy.mesh = buoy.mesh.keep_immersed_part()
    buoy.add_translation_dof(name="Sway")

    two_buoys = FloatingBody.join_bodies(buoy, buoy.translated_x(distance))
    two_buoys.mesh = buoy.mesh.symmetrized(yOz_Plane)  # Use a ReflectionSymmetry as mesh...

    problems = [RadiationProblem(body=two_buoys, omega=1.0, radiating_dof=dof) for dof in two_buoys.dofs]

    results = assemble_dataset(solver_without_sym.solve_all(problems))
    # Check that the resulting matrix is symmetric
    assert np.isclose(results['added_mass'].data[0, 0, 0], results['added_mass'].data[0, 1, 1])
    assert np.isclose(results['added_mass'].data[0, 1, 0], results['added_mass'].data[0, 0, 1])
    assert np.isclose(results['radiation_damping'].data[0, 0, 0], results['radiation_damping'].data[0, 1, 1])
    assert np.isclose(results['radiation_damping'].data[0, 1, 0], results['radiation_damping'].data[0, 0, 1])

    results_with_sym = assemble_dataset(solver_with_sym.solve_all(problems))
    assert np.allclose(results['added_mass'].data, results_with_sym['added_mass'].data, rtol=1e-2)
    assert np.allclose(results['radiation_damping'].data, results_with_sym['radiation_damping'].data, rtol=1e-3)


@pytest.mark.parametrize("method", method)
def test_odd_axial_symmetry(method):
    """Buoy with odd number of slices."""
    def shape(z):
            return 0.1*(-(z+1)**2 + 16)
    buoy = FloatingBody(AxialSymmetricMesh.from_profile(shape, z_range=np.linspace(-5.0, 0.0, 9), nphi=5))
    buoy.add_translation_dof(direction=(0, 0, 1), name="Heave")

    problem = RadiationProblem(body=buoy, omega=2.0)
    result1 = solver_with_sym.solve(problem,method=method)

    full_buoy = FloatingBody(buoy.mesh.merged())
    full_buoy.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=full_buoy, omega=2.0)
    result2 = solver_with_sym.solve(problem,method=method)

    volume = buoy.mesh.volume
    assert np.isclose(result1.added_masses["Heave"], result2.added_masses["Heave"], atol=1e-4*volume*problem.rho)
    assert np.isclose(result1.radiation_dampings["Heave"], result2.radiation_dampings["Heave"], atol=1e-4*volume*problem.rho)


@pytest.mark.parametrize("method", method)
@pytest.mark.parametrize("depth", [10.0, np.inf])
def test_horizontal_cylinder(depth,method):
    cylinder = HorizontalCylinder(length=10.0, radius=1.0, reflection_symmetry=False, translation_symmetry=False, nr=2, ntheta=10, nx=10)
    assert isinstance(cylinder.mesh, Mesh)
    cylinder.translate_z(-3.0)
    cylinder.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=cylinder, omega=1.0, water_depth=depth)
    result1 = solver_with_sym.solve(problem,method=method)

    trans_cylinder = HorizontalCylinder(length=10.0, radius=1.0, reflection_symmetry=False, translation_symmetry=True, nr=2, ntheta=10, nx=10)
    assert isinstance(trans_cylinder.mesh, CollectionOfMeshes)
    assert isinstance(trans_cylinder.mesh[0], TranslationalSymmetricMesh)
    trans_cylinder.translate_z(-3.0)
    trans_cylinder.add_translation_dof(direction=(0, 0, 1), name="Heave")
    problem = RadiationProblem(body=trans_cylinder, omega=1.0, water_depth=depth)
    result2 = solver_with_sym.solve(problem,method=method)

    # S, V = solver_with_sym.build_matrices(trans_cylinder.mesh, trans_cylinder.mesh)
    # S.plot_shape()

    assert np.isclose(result1.added_masses["Heave"], result2.added_masses["Heave"], atol=1e-4*cylinder.volume*problem.rho)
    assert np.isclose(result1.radiation_dampings["Heave"], result2.radiation_dampings["Heave"], atol=1e-4*cylinder.volume*problem.rho)


# HIERARCHICAL MATRICES
@pytest.mark.parametrize("method", method)
def test_low_rank_matrices(method):
    radius = 1.0
    resolution = 2
    perimeter = 2*np.pi*radius
    buoy = Sphere(radius=radius, center=(0.0, 0.0, 0.0),
                  ntheta=int(perimeter*resolution/2), nphi=int(perimeter*resolution),
                  clip_free_surface=True, axial_symmetry=False, name=f"buoy")
    buoy.add_translation_dof(name="Heave")
    two_distant_buoys = FloatingBody.join_bodies(buoy, buoy.translated_x(20))
    two_distant_buoys.mesh._meshes[1].name = "other_buoy_mesh"

    S, V = solver_with_sym.engine.build_matrices(two_distant_buoys.mesh, two_distant_buoys.mesh, 0.0, np.inf, 1.0, solver_with_sym.green_function)
    assert isinstance(S.all_blocks[0, 1], LowRankMatrix)
    assert isinstance(S.all_blocks[1, 0], LowRankMatrix)
    # S.plot_shape()

    problem = RadiationProblem(body=two_distant_buoys, omega=1.0, radiating_dof="buoy__Heave")
    result = solver_with_sym.solve(problem,method=method)
    result2 = solver_without_sym.solve(problem,method=method)

    assert np.isclose(result.added_masses['buoy__Heave'], result2.added_masses['buoy__Heave'], atol=10.0)
    assert np.isclose(result.radiation_dampings['buoy__Heave'], result2.radiation_dampings['buoy__Heave'], atol=10.0)


@pytest.mark.parametrize("method,adjoint_double_layer",zip(method,adjoint_double_layer))
def test_array_of_spheres(method,adjoint_double_layer):
    radius = 1.0
    resolution = 2
    perimeter = 2*np.pi*radius
    buoy = Sphere(radius=radius, center=(0.0, 0.0, 0.0),
                  ntheta=int(perimeter*resolution/2), nphi=int(perimeter*resolution),
                  clip_free_surface=True, axial_symmetry=False, name=f"buoy")
    buoy.add_translation_dof(name="Surge")
    buoy.add_translation_dof(name="Sway")
    buoy.add_translation_dof(name="Heave")

    # Corner case
    dumb_array = buoy.assemble_regular_array(distance=5.0, nb_bodies=(1, 1))
    assert dumb_array.mesh == buoy.mesh

    # Main case
    array = buoy.assemble_regular_array(distance=4.0, nb_bodies=(3, 3))

    assert isinstance(array.mesh, TranslationalSymmetricMesh)
    assert isinstance(array.mesh[0], TranslationalSymmetricMesh)
    assert array.mesh[0][0] == buoy.mesh

    assert len(array.dofs) == 3*3*3
    assert "2_0__Heave" in array.dofs

    #
    array = buoy.assemble_regular_array(distance=4.0, nb_bodies=(3, 1))

    fullS, fullV = solver_without_sym.engine.build_matrices(array.mesh, array.mesh, 0.0, np.inf,
			1.0, solver_without_sym.green_function, adjoint_double_layer=adjoint_double_layer)
    S, V = solver_with_sym.engine.build_matrices(array.mesh, array.mesh, 0.0, np.inf,
			1.0, solver_with_sym.green_function, adjoint_double_layer=adjoint_double_layer)

    assert isinstance(S, cpt.matrices.block.BlockMatrix)
    assert np.allclose(S.full_matrix(), fullS)
    assert np.allclose(V.full_matrix(), fullV)

    problem = RadiationProblem(body=array, omega=1.0, radiating_dof="2_0__Heave", water_depth=np.inf)

    result = solver_with_sym.solve(problem, method=method)
    result2 = solver_without_sym.solve(problem, method=method)

    assert np.isclose(result.added_masses['2_0__Heave'], result2.added_masses['2_0__Heave'], atol=15.0)
    assert np.isclose(result.radiation_dampings['2_0__Heave'], result2.radiation_dampings['2_0__Heave'], atol=15.0)



"""Variants of Delhommeau's method for the computation of the Green function."""
# Copyright (C) 2017-2024 Matthieu Ancellin
# See LICENSE file at <https://github.com/capytaine/capytaine>

import os
import logging
from functools import lru_cache
from importlib import import_module

import numpy as np

from capytaine.meshes.meshes import Mesh
from capytaine.meshes.collections import CollectionOfMeshes
from capytaine.tools.prony_decomposition import exponential_decomposition, error_exponential_decomposition
from capytaine.tools.cache_on_disk import cache_directory

from capytaine.green_functions.abstract_green_function import AbstractGreenFunction

LOG = logging.getLogger(__name__)

_default_parameters = dict(
    tabulation_nr=676,
    tabulation_rmax=100.0,
    tabulation_nz=372,
    tabulation_zmin=-251.0,
    tabulation_nb_integration_points=1000,
    tabulation_method="scaled_nemoh3",
    finite_depth_prony_decomposition_method="fortran",
    floating_point_precision="float64"
)


class Delhommeau(AbstractGreenFunction):
    """The Green function as implemented in Aquadyn and Nemoh.

    Parameters
    ----------
    tabulation_nr: int, optional
        Number of tabulation points for horizontal coordinate.
        If 0 is given, no tabulation is used at all.
        Default: 676
    tabulation_rmax: float, optional
        Maximum value of r range for the tabulation. (Minimum is zero.)
        Only used with the :code:`"scaled_nemoh3"` method.
        Default: 100.0
    tabulation_nz: int, optional
        Number of tabulation points for vertical coordinate.
        If 0 is given, no tabulation is used at all.
        Default: 372
    tabulation_zmin: float, optional
        Minimum value of z range for the tabulation. (Maximum is zero.)
        Only used with the :code:`"scaled_nemoh3"` method.
        Default: -251.0
    tabulation_nb_integration_points: int, optional
        Number of points for the numerical integration w.r.t. :math:`theta` of
        Delhommeau's integrals
        Default: 1000
    tabulation_method: string, optional
        Either :code:`"legacy"` or :code:`"scaled_nemoh3"`, which are the two
        methods currently implemented.
        Default: :code:`"scaled_nemoh3"`
    finite_depth_prony_decomposition_method: string, optional
        The implementation of the Prony decomposition used to compute the
        finite water_depth Green function. Accepted values: :code:`'fortran'`
        for Nemoh's implementation (by default), :code:`'python'` for an
        experimental Python implementation.
        See :func:`find_best_exponential_decomposition`.
    floating_point_precision: string, optional
        Either :code:`'float32'` for single precision computations or
        :code:`'float64'` for double precision computations.
        Default: :code:`'float64'`.

    Attributes
    ----------
    fortran_core:
        Compiled Fortran module with functions used to compute the Green
        function.
    tabulation_method_index: int
        Integer passed to Fortran code to describe which method is used.
    tabulated_r_range: numpy.array of shape (tabulation_nr,) and type floating_point_precision
    tabulated_z_range: numpy.array of shape (tabulation_nz,) and type floating_point_precision
        Coordinates of the tabulation points.
    tabulated_integrals: numpy.array of shape (tabulation_nr, tabulation_nz, 2, 2) and type floating_point_precision
        Tabulated Delhommeau integrals.
    """

    fortran_core_basename = "Delhommeau"


    def __init__(self, *,
                 tabulation_nr=_default_parameters["tabulation_nr"],
                 tabulation_rmax=_default_parameters["tabulation_rmax"],
                 tabulation_nz=_default_parameters["tabulation_nz"],
                 tabulation_zmin=_default_parameters["tabulation_zmin"],
                 tabulation_nb_integration_points=_default_parameters["tabulation_nb_integration_points"],
                 tabulation_method=_default_parameters["tabulation_method"],
                 finite_depth_prony_decomposition_method=_default_parameters["finite_depth_prony_decomposition_method"],
                 floating_point_precision=_default_parameters["floating_point_precision"],
                 ):

        self.floating_point_precision = floating_point_precision

        self.fortran_core = import_module(f"capytaine.green_functions.libs.{self.fortran_core_basename}_{self.floating_point_precision}")

        self.tabulation_method = tabulation_method
        fortran_indices_for_methods = {
                'legacy': self.fortran_core.delhommeau_integrals.legacy_method,
                'scaled_nemoh3': self.fortran_core.delhommeau_integrals.scaled_nemoh3_method,
                              }
        self.tabulation_method_index = fortran_indices_for_methods[tabulation_method]

        self._create_or_load_tabulation(tabulation_nr, tabulation_rmax, tabulation_nz, tabulation_zmin, tabulation_nb_integration_points)

        self.finite_depth_prony_decomposition_method = finite_depth_prony_decomposition_method

        self.exportable_settings = {
            'green_function': self.__class__.__name__,
            'tabulation_nr': tabulation_nr,
            'tabulation_rmax': tabulation_rmax,
            'tabulation_nz': tabulation_nz,
            'tabulation_zmin': tabulation_zmin,
            'tabulation_nb_integration_points': tabulation_nb_integration_points,
            'tabulation_method': tabulation_method,
            'finite_depth_prony_decomposition_method': finite_depth_prony_decomposition_method,
            'floating_point_precision': floating_point_precision,
        }

        self._hash = hash(self.exportable_settings.values())

    def __hash__(self):
        return self._hash

    def __str__(self):
        # Print only the non-default values.
        to_be_printed = []
        for name, value in self.exportable_settings.items():
            if name in _default_parameters and value != _default_parameters[name]:
                to_be_printed.append(f"{name}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(to_be_printed)})"

    def __repr__(self):
        # Same as __str__ except all values are printed even when they are the
        # default value.
        to_be_printed = []
        for name, value in self.exportable_settings.items():
            if name in _default_parameters:
                to_be_printed.append(f"{name}={repr(value)}")
        return f"{self.__class__.__name__}({', '.join(to_be_printed)})"

    def _repr_pretty_(self, p, cycle):
        p.text(self.__repr__())

    def _create_or_load_tabulation(self, tabulation_nr, tabulation_rmax,
                                   tabulation_nz, tabulation_zmin,
                                   tabulation_nb_integration_points):
        """This method either:
            - loads an existing tabulation saved on disk
            - generates a new tabulation with the data provided as argument and save it on disk.
        """

        # Normalize inputs
        tabulation_rmax = float(tabulation_rmax)
        tabulation_zmin = float(tabulation_zmin)

        filename = "tabulation_{}_{}_{}_{}_{}_{}_{}_{}.npz".format(
            self.fortran_core_basename, self.floating_point_precision,
            self.tabulation_method,
            tabulation_nr, tabulation_rmax, tabulation_nz, tabulation_zmin,
            tabulation_nb_integration_points
        )
        filepath = os.path.join(cache_directory(), filename)

        if os.path.exists(filepath):
            LOG.info("Loading tabulation from %s", filepath)
            loaded_arrays = np.load(filepath)
            self.tabulated_r_range = loaded_arrays["r_range"]
            self.tabulated_z_range = loaded_arrays["z_range"]
            self.tabulated_integrals = loaded_arrays["values"]

        else:
            LOG.warning("Precomputing tabulation, it may take a few seconds.")
            self.tabulated_r_range = self.fortran_core.delhommeau_integrals.default_r_spacing(
                    tabulation_nr, tabulation_rmax, self.tabulation_method_index
                    )
            self.tabulated_z_range = self.fortran_core.delhommeau_integrals.default_z_spacing(
                    tabulation_nz, tabulation_zmin, self.tabulation_method_index
                    )
            self.tabulated_integrals = self.fortran_core.delhommeau_integrals.construct_tabulation(
                    self.tabulated_r_range, self.tabulated_z_range, tabulation_nb_integration_points
                    )
            LOG.debug("Saving tabulation in %s", filepath)
            np.savez_compressed(
                filepath, r_range=self.tabulated_r_range, z_range=self.tabulated_z_range,
                values=self.tabulated_integrals
            )

    @lru_cache(maxsize=128)
    def find_best_exponential_decomposition(self, dimensionless_omega, dimensionless_wavenumber):
        """Compute the decomposition of a part of the finite water_depth Green function as a sum of exponential functions.

        Two implementations are available: the legacy Fortran implementation from Nemoh and a newer one written in Python.
        For some still unexplained reasons, the two implementations do not always give the exact same result.
        Until the problem is better understood, the Fortran implementation is the default one, to ensure consistency with Nemoh.
        The Fortran version is also significantly faster...

        Results are cached.

        Parameters
        ----------
        dimensionless_omega: float
            dimensionless angular frequency: :math:`kh \\tanh (kh) = \\omega^2 h/g`
        dimensionless_wavenumber: float
            dimensionless wavenumber: :math:`kh`
        method: string, optional
            the implementation that should be used to compute the Prony decomposition

        Returns
        -------
        Tuple[np.ndarray, np.ndarray]
            the amplitude and growth rates of the exponentials
        """

        LOG.debug(f"\tCompute Prony decomposition in finite water_depth Green function "
                  f"for dimless_omega=%.2e and dimless_wavenumber=%.2e",
                  dimensionless_omega, dimensionless_wavenumber)

        if self.finite_depth_prony_decomposition_method.lower() == 'python':
            # The function that will be approximated.
            @np.vectorize
            def f(x):
                return self.fortran_core.initialize_green_wave.ff(x, dimensionless_omega, dimensionless_wavenumber)

            # Try different increasing number of exponentials
            for n_exp in range(4, 31, 2):

                # The coefficients are computed on a resolution of 4*n_exp+1 ...
                X = np.linspace(-0.1, 20.0, 4*n_exp+1)
                a, lamda = exponential_decomposition(X, f(X), n_exp)

                # ... and they are evaluated on a finer discretization.
                X = np.linspace(-0.1, 20.0, 8*n_exp+1)
                if error_exponential_decomposition(X, f(X), a, lamda) < 1e-4:
                    break

            else:
                LOG.warning("No suitable exponential decomposition has been found"
                            "for dimless_omega=%.2e and dimless_wavenumber=%.2e",
                            dimensionless_omega, dimensionless_wavenumber)

        elif self.finite_depth_prony_decomposition_method.lower() == 'fortran':
            lamda, a, nexp = self.fortran_core.old_prony_decomposition.lisc(dimensionless_omega, dimensionless_wavenumber)
            lamda = lamda[:nexp]
            a = a[:nexp]

        else:
            raise ValueError("Unrecognized method name for the Prony decomposition.")

        # Add one more exponential function (actually a constant).
        # It is not clear where it comes from exactly in the theory...
        a = np.concatenate([a, np.array([2])])
        lamda = np.concatenate([lamda, np.array([0.0])])

        return a, lamda

    def evaluate(self, mesh1, mesh2, free_surface=0.0, water_depth=np.inf, wavenumber=1.0, adjoint_double_layer=True, early_dot_product=True):
        r"""The main method of the class, called by the engine to assemble the influence matrices.

        Parameters
        ----------
        mesh1: Mesh or CollectionOfMeshes or list of points
            mesh of the receiving body (where the potential is measured)
            if only S is wanted or early_dot_product is False, then only a list of points as an array of shape (n, 3) can be passed.
        mesh2: Mesh or CollectionOfMeshes
            mesh of the source body (over which the source distribution is integrated)
        free_surface: float, optional
            position of the free surface (default: :math:`z = 0`)
        water_depth: float, optional
            constant depth of water (default: :math:`+\infty`)
        wavenumber: float, optional
            wavenumber (default: 1.0)
        adjoint_double_layer: bool, optional
            compute double layer for direct method (F) or adjoint double layer for indirect method (T) matrices (default: True)
        early_dot_product: boolean, optional
            if False, return K as a (n, m, 3) array storing ∫∇G
            if True, return K as a (n, m) array storing ∫∇G·n

        Returns
        -------
        tuple of numpy arrays
            the matrices :math:`S` and :math:`K`
        """

        wavenumber = float(wavenumber)

        if free_surface == np.inf: # No free surface, only a single Rankine source term

            a_exp, lamda_exp = np.empty(1), np.empty(1)  # Dummy arrays that won't actually be used by the fortran code.

            coeffs = np.array((1.0, 0.0, 0.0))

        elif water_depth == np.inf:

            a_exp, lamda_exp = np.empty(1), np.empty(1)  # Idem

            if wavenumber == 0.0:
                coeffs = np.array((1.0, 1.0, 0.0))
            elif wavenumber == np.inf:
                coeffs = np.array((1.0, -1.0, 0.0))
            else:
                coeffs = np.array((1.0, 1.0, 1.0))

        else:  # Finite water_depth
            if wavenumber == 0.0 or wavenumber == np.inf:
                raise NotImplementedError("Zero or infinite frequencies not implemented for finite depth.")
            else:
                a_exp, lamda_exp = self.find_best_exponential_decomposition(
                    wavenumber*water_depth*np.tanh(wavenumber*water_depth),
                    wavenumber*water_depth,
                )
                coeffs = np.array((1.0, 1.0, 1.0))

        if isinstance(mesh1, Mesh) or isinstance(mesh1, CollectionOfMeshes):
            collocation_points = mesh1.faces_centers
            nb_collocation_points = mesh1.nb_faces
            if ( adjoint_double_layer == False ):
                early_dot_product_normals = np.zeros((nb_collocation_points, 3))  # Should not be used
            else:
                early_dot_product_normals = mesh1.faces_normals
        elif isinstance(mesh1, np.ndarray) and mesh1.ndim ==2 and mesh1.shape[1] == 3:
            # This is used when computing potential or velocity at given points in postprocessing
            collocation_points = mesh1
            nb_collocation_points = mesh1.shape[0]
            early_dot_product_normals = np.zeros((nb_collocation_points, 3))  # Should not be used
            if ( adjoint_double_layer == False ):
                raise NotImplementedError("Using a list of points as collocation points is not supported in computing adjoint double layer matrices.")
        else:
            raise ValueError(f"Unrecognized input for {self.__class__.__name__}.evaluate")

        if self.floating_point_precision == "float32":
            dtype = "complex64"
        elif self.floating_point_precision == "float64":
            dtype = "complex128"
        else:
            raise NotImplementedError

        S = np.empty((nb_collocation_points, mesh2.nb_faces), order="F", dtype=dtype)
        K = np.empty((nb_collocation_points, mesh2.nb_faces, 1 if early_dot_product else 3), order="F", dtype=dtype)

        # Main call to Fortran code
        self.fortran_core.matrices.build_matrices(
            collocation_points,  early_dot_product_normals,
            mesh2.vertices,      mesh2.faces + 1,
            mesh2.faces_centers, mesh2.faces_normals,
            mesh2.faces_areas,   mesh2.faces_radiuses,
            *mesh2.quadrature_points,
            wavenumber, water_depth,
            coeffs,
            self.tabulation_method_index, self.tabulated_r_range, self.tabulated_z_range, self.tabulated_integrals,
            lamda_exp, a_exp,
            mesh1 is mesh2, adjoint_double_layer,
            S, K
        )

        if np.any(np.isnan(S)) or np.any(np.isnan(K)):
            raise RuntimeError("Green function returned a NaN in the interaction matrix.\n"
                    "It could be due to overlapping panels.")

        if early_dot_product: K = K.reshape((nb_collocation_points, mesh2.nb_faces))

        return S, K

################################

class XieDelhommeau(Delhommeau):
    """Variant of Nemoh's Green function, more accurate near the free surface.

    Same arguments and methods as :class:`Delhommeau`.
    """

    fortran_core_basename = "XieDelhommeau"



from functools import lru_cache

import pytest
import numpy as np

from capytaine.tools.lru_cache import lru_cache_with_strict_maxsize


def get_random_id(n):
    # An impure function that shouldn't be memoized, but allows to test the cache.
    return np.random.default_rng().integers(n)

N = 100_000_000

def test_cache():
    f = lru_cache_with_strict_maxsize()(get_random_id)
    a = f(N)
    b = f(N)
    assert a == b  # randint was not run the second time, the cached value was used


def test_cache_maxsize():
    # 1/N chance of failing
    f = lru_cache_with_strict_maxsize(maxsize=1)(get_random_id)
    a = f(N)
    f(10)
    b = f(N)
    assert a != b  # the cached value of the first call had been deleted


def test_cache_kwargs():
    f = lru_cache_with_strict_maxsize()(get_random_id)
    a = f(n=N)
    b = f(n=N)
    assert a == b  # the cached value of the first call had been deleted


def test_delete_first():
    """This test is a convoluted way to test the main difference between the
    built-in lru_cache and the lru_cache_with_strict_maxsize decorator defined
    by Capytaine. The difference is that the latter make sure that objects in
    the cache are deleted before computing and caching a new result. For
    Capytaine, it is meant to limit the RAM usage. Here, it is tested by making
    sure that there is never two instances of a Singleton class at the same
    time."""

    class Singleton:
        nb_instances = [0]
        # Use a one-element list as a mutable integer shared by all instances.

        def __init__(self):
            if self.nb_instances[0] != 0:
                raise ValueError("There can be only one Singleton at the same time!")
            self.nb_instances[0] += 1

        def __del__(self):
            # When the singleton is deleted, update the counter
            self.nb_instances[0] -= 1

    assert Singleton.nb_instances[0] == 0
    a = Singleton()
    assert Singleton.nb_instances[0] == 1
    del a
    assert Singleton.nb_instances[0] == 0

    ## NO CACHE
    def new_instance(a):
        return Singleton()
    new_instance(1)
    assert Singleton.nb_instances[0] == 0  # The Singleton created above is immediately garbage-collected.

    ## CAPYTAINE'S CACHE
    @lru_cache_with_strict_maxsize(maxsize=1)
    def new_instance(a):
        return Singleton()

    new_instance(1)
    assert Singleton.nb_instances[0] == 1  # It is not garbage collected only because it is still in the cache.
    new_instance(2)  # When we create a new instance, the old one is removed from the cache before creating the new one.

    del new_instance  # Delete the cache before doing more tests.

    # STDLIB CACHE
    @lru_cache(maxsize=1)
    def new_instance(a):
        return Singleton()

    new_instance(1)
    with pytest.raises(ValueError):
        new_instance(2)  # lru_cache tries to create the new singleton, before deleting the old one from the cache.


# def benchmark_ram_usage():
#     # Would need a way to monitor the RAM usage to automate this test
#     # The RAM usage should never be more than the ~3GB allocated by a single call
#
#     # @delete_first_lru_cache(maxsize=1)
#     @lru_cache(maxsize=1)
#     def f(i):
#         np.random.seed(i)
#         a = np.random.random((20_000, 20_000))
#         print(f"Allocated {a.nbytes / 2**30:.2f} GB")
#         return a
#
#     for i in (1, 1, 2, 1, 3):
#         print(f"Calling f({i})")
#         f(i)



#!/usr/bin/env python
# coding: utf-8
"""Test of the matrices submodule."""

from warnings import warn
import pytest

import numpy as np
from numpy.linalg import norm, matrix_rank

from capytaine.matrices.block import BlockMatrix
from capytaine.matrices.block_toeplitz import *
from capytaine.matrices.builders import *
from capytaine.matrices.low_rank import LowRankMatrix, NoConvergenceOfACA

try:
    import matplotlib.pyplot as plt
except ImportError:
    plt = None

@pytest.fixture
def two_by_two_block_identity():
    A = BlockMatrix([
        [np.eye(2, 2), np.zeros((2, 2))],
        [np.zeros((2, 2)), np.eye(2, 2)]
    ])
    return A


def test_block_matrix_representation_of_identity_two_by_two(two_by_two_block_identity):
    # 2x2 block representation of the identity matrix
    A = two_by_two_block_identity
    assert A.shape == (4, 4)
    assert A.nb_blocks == (2, 2)
    assert A.block_shapes == ([2, 2], [2, 2])
    assert list(A._stored_block_positions()) == [[(0, 0)], [(0, 2)], [(2, 0)], [(2, 2)]]
    assert A.str_shape == "2×2×[2×2]"
    assert A.density == 1.0

    assert ((A + A)/2 == A).all()
    assert (-A).min() == -1
    assert (2*A).max() == 2
    assert (A*A == A).all()
    b = np.random.rand(4)
    assert (A @ b == b).all()
    assert (A @ A == A).all()

    assert (A == identity_like(A)).all()
    assert (A.full_matrix() == np.eye(4, 4)).all()
    assert (np.array(A) == A.full_matrix()).all()


@pytest.mark.skipif(plt is None,
                     reason='matplotlib is not installed')
def test_block_matrix_representation_of_identity_patches(two_by_two_block_identity):
    A = two_by_two_block_identity
    patches = A._patches(global_frame=(10, 10))
    exp = {(10, 10), (12, 10), (10, 12), (12, 12)}
    assert {rectangle.get_xy() for rectangle in patches} == exp


@pytest.fixture
def block_three_rect(two_by_two_block_identity):
    A = two_by_two_block_identity
    B = BlockMatrix([[A, np.zeros((4, 1))], [np.zeros((1, 4)), np.ones((1, 1))]])
    return B


def test_block_matrix_representation_of_identity_rect(block_three_rect):
    # 2x2 block matrix containing one another block matrix and three regular matrices of different shapes
    B = block_three_rect
    assert B.shape == (5, 5)
    assert B.block_shapes == ([4, 1], [4, 1])
    assert (B.full_matrix() == np.eye(5, 5)).all()
    assert B.str_shape == "[[2×2×[2×2], 4×1], [1×4, 1×1]]"
    assert B.block_shapes == ([4, 1], [4, 1])
    assert list(B._stored_block_positions()) == [[(0, 0)], [(0, 4)], [(4, 0)], [(4, 4)]]


@pytest.mark.skipif(plt is None,
                     reason='matplotlib is not installed')
def test_block_matrix_representation_of_identity_rect_patch(block_three_rect):
    B = block_three_rect
    patches = B._patches(global_frame=(10, 10))
    exp = {(10, 10), (12, 10), (10, 12), (12, 12), (14, 10), (10, 14), (14, 14)}
    assert {rectangle.get_xy() for rectangle in patches} == exp

def test_block_diff_shapes():
    # 3x3 block matrix with blocks of different shapes
    C = random_block_matrix([1, 2, 4], [1, 2, 4])
    assert C.nb_blocks == (3, 3)
    assert C.block_shapes == ([1, 2, 4], [1, 2, 4])
    assert (ones_like(C).full_matrix() ==  np.ones(C.shape)).all()

    assert (cut_matrix(C.full_matrix(), *C.block_shapes) == C).all()

    assert (C @ random_block_matrix(C.block_shapes[1], [6])).block_shapes == ([1, 2, 4], [6])

    b = np.random.rand(7)
    assert np.allclose(C @ b, C.full_matrix() @ b)



def test_sparse_storage_of_block_toeplitz_matrices():
    A = BlockToeplitzMatrix(
        [[np.array([[i]]) for i in range(5)]]
    )
    assert np.all(A.full_matrix() ==
                  np.array([[0, 1, 2],
                            [4, 0, 1],
                            [3, 4, 0]]))
    assert A.density == 5 / 9

    assert not isinstance(A.no_toeplitz(), BlockToeplitzMatrix)
    assert np.all(A.no_toeplitz().full_matrix() == A.full_matrix())

    B = BlockSymmetricToeplitzMatrix(
        [[np.array([[i]]) for i in range(4)]]
    )
    assert np.all(B.full_matrix() ==
                  np.array([[0, 1, 2, 3],
                            [1, 0, 1, 2],
                            [2, 1, 0, 1],
                            [3, 2, 1, 0],
                            ]))
    assert B.density == 4 / 16

    C = BlockCirculantMatrix(
        [[np.array([[i]]) for i in range(4)]]
    )
    assert np.all(C.full_matrix() ==
                  np.array([[0, 1, 2, 3],
                            [3, 0, 1, 2],
                            [2, 3, 0, 1],
                            [1, 2, 3, 0],
                            ]))
    assert C.density == 4 / 16

    D = EvenBlockSymmetricCirculantMatrix(
        [[np.array([[i]]) for i in range(3)]]
    )
    assert np.all(D.full_matrix() ==
                  np.array([[0, 1, 2, 1],
                            [1, 0, 1, 2],
                            [2, 1, 0, 1],
                            [1, 2, 1, 0],
                            ]))
    assert D.density == 3 / 16

    E = OddBlockSymmetricCirculantMatrix(
        [[np.array([[i]]) for i in range(3)]]
    )
    assert np.all(E.full_matrix() ==
                  np.array([[0, 1, 2, 2, 1],
                            [1, 0, 1, 2, 2],
                            [2, 1, 0, 1, 2],
                            [2, 2, 1, 0, 1],
                            [1, 2, 2, 1, 0],
                            ]))
    assert E.density == 3 / 25


def test_block_toeplitz_matrices():
    # 2x2 block Toeplitz representation of the identity matrix
    A = BlockToeplitzMatrix([
        [np.eye(2, 2), np.zeros((2, 2)), np.zeros((2, 2))]
    ])
    assert A.block_shapes == ([2, 2], [2, 2])
    assert A.block_shape == (2, 2)
    assert A.shape == (4, 4)
    assert list(A._stored_block_positions()) == [[(0, 0), (2, 2)], [(0, 2)], [(2, 0)]]
    assert (A.full_matrix() == np.eye(*A.shape)).all()

    # Removing the Toeplitz structure.
    Ant = A.no_toeplitz()
    assert not isinstance(Ant, BlockToeplitzMatrix)
    assert Ant.all_blocks[0, 0] is Ant.all_blocks[1, 1]
    # When using no_toeplitz, the resulting BlockMatrix still contains several references to the same array object.

    from copy import deepcopy
    Ant_cp = deepcopy(Ant)
    assert Ant_cp.all_blocks[0, 0] is not Ant_cp.all_blocks[1, 1]
    # The deepcopy has made different copies of the blocks in the block matrix without explicit Toeplitz structure.

    # However,
    A_cp = deepcopy(A)
    assert A_cp.all_blocks[0, 0] is A_cp.all_blocks[1, 1]
    # The matrix with explicit Toeplitz structure still keep the structure when deepcopied.

    assert ((A + A)/2 == A).all()
    assert (-A).min() == -1
    assert (2*A).max() == 2
    assert (A*A == A).all()

    assert isinstance(A.circulant_super_matrix, BlockCirculantMatrix)
    assert A.circulant_super_matrix.shape == (6, 6)

    b = np.random.rand(A.shape[0])
    assert np.allclose(A @ b, b)
    assert np.allclose(A.rmatvec(b), b)

    # 3x3 block random Toeplitz matrix
    B = BlockToeplitzMatrix([
        [
            random_block_matrix([1], [1]),
            random_block_matrix([1], [1]),
            random_block_matrix([1], [1]),
            random_block_matrix([1], [1]),
            random_block_matrix([1], [1]),
        ]
    ])
    assert B.shape == (3, 3)
    assert B.nb_blocks == (3, 3)
    assert B.block_shape == (1, 1)
    assert list(B._stored_block_positions()) == [[(0, 0), (1, 1), (2, 2)], [(0, 1), (1, 2)], [(0, 2)], [(2, 0)], [(1, 0), (2, 1)]]

    assert isinstance(B.circulant_super_matrix, BlockCirculantMatrix)
    assert B.circulant_super_matrix.shape == (5, 5)

    b = np.random.rand(B.shape[0])
    assert np.allclose(BlockMatrix.matvec(B, b), B.full_matrix() @ b)
    assert np.allclose(B @ b, B.full_matrix() @ b)
    assert np.allclose(B.rmatvec(b), b @ B.full_matrix())


def test_block_symmetric_toeplitz_matrices():
    # 2x2 block symmetric Toeplitz representation of the identity matrix
    A = BlockSymmetricToeplitzMatrix([
        [np.eye(2, 2), np.zeros((2, 2))]
    ])
    assert A.nb_blocks == (2, 2)
    assert A.block_shapes == ([2, 2], [2, 2])
    assert A.shape == (4, 4)
    assert A.block_shape == (2, 2)
    assert list(A._stored_block_positions()) == [[(0, 0), (2, 2)], [(0, 2), (2, 0)]]
    assert (A.full_matrix() == np.eye(*A.shape)).all()

    assert ((A + A)/2 == A).all()
    assert (-A).min() == -1
    assert (2*A).max() == 2
    assert (A*A == A).all()

    b = np.random.rand(4)
    assert np.allclose(A @ b, b)
    assert np.allclose(A.rmatvec(b), b)

    # The same as a simpler block matrix
    A2 = BlockMatrix(A.all_blocks)
    assert (A2.full_matrix() == A.full_matrix()).all()

    # 2x2 blocks symmetric Toeplitz matrix of non-squared matrix
    A3 = BlockSymmetricToeplitzMatrix([
        [np.ones((3, 1)), np.zeros((3, 1))]
    ])
    assert A3.nb_blocks == (2, 2)
    assert A3.block_shapes == ([3, 3], [1, 1])
    assert A3.shape == (6, 2)
    assert A3.block_shape == (3, 1)

    with pytest.raises(AssertionError):
        BlockSymmetricToeplitzMatrix([
            [np.ones((1, 1)), np.zeros((2, 2))]
        ])

    # 2x2 block symmetrix Toeplitz matrix composed of block matrices
    B = BlockSymmetricToeplitzMatrix([
        [random_block_matrix([1, 1], [1, 1]), random_block_matrix([1, 1], [1, 1])]
    ])
    assert B.nb_blocks == (2, 2)
    assert B._stored_nb_blocks == (1, 2)
    assert B.block_shapes == ([2, 2], [2, 2])
    assert B.block_shape == (2, 2)
    assert B.shape == (4, 4)

    assert isinstance(B.all_blocks[0, 0], BlockMatrix)
    assert (B.all_blocks[0, 0] == B.all_blocks[1, 1]).all()

    list_of_matrices = [B, B + ones_like(B), B, B - ones_like(B)]
    BB_fft = BlockSymmetricToeplitzMatrix.fft_of_list(*list_of_matrices)
    full_BB_fft = np.fft.fft(np.array([A.full_matrix() for A in list_of_matrices]), axis=0)
    assert np.allclose(full_BB_fft, np.array([A.full_matrix() for A in BB_fft]))

    # 2x2 block symmetric Toeplitz matrix composed of the previous block symmetric Toeplitz matrices
    C = BlockSymmetricToeplitzMatrix([
        [A, B]
    ])
    assert C.shape == (8, 8)

    b = np.random.rand(8)
    assert np.allclose(C @ b, C.full_matrix() @ b)
    assert np.allclose(C.rmatvec(b), b @ C.full_matrix())

    # We need to go deeper
    D = BlockMatrix([
        [C, np.zeros(C.shape)]
    ])
    assert D.shape == (8, 16)

    b = np.random.rand(16)
    assert np.allclose(D @ b, D.full_matrix() @ b)


def test_block_circulant_matrix():
    # 5x5 block symmetric circulant matrix representation of the identity matrix
    A = BlockCirculantMatrix([
        [np.eye(2, 2), np.zeros((2, 2)), np.zeros((2, 2))]
    ])
    assert A.nb_blocks == (3, 3)
    assert A.shape == (6, 6)
    assert A.block_shapes == ([2, 2, 2], [2, 2, 2])
    assert (A.full_matrix() == np.eye(*A.shape)).all()

    assert ((A + A)/2 == A).all()
    assert (-A).min() == -1
    assert (2*A).max() == 2
    assert (A*A == A).all()

    b = np.random.rand(A.shape[0])
    assert np.allclose(A @ b, A.full_matrix() @ b)
    assert np.allclose(A.rmatvec(b), b @ A.full_matrix())

    # Nested matrix
    B = BlockCirculantMatrix([[A, 2*A, 3*A]])
    assert B.nb_blocks == (3, 3)
    assert B.shape == (18, 18)
    assert (B.all_blocks[0, 0] == A).all()

    C = BlockCirculantMatrix([[np.array([[i]]) for i in range(5)]])
    b = np.random.rand(C.shape[0])
    assert np.allclose(BlockMatrix.matvec(C, b), C.full_matrix() @ b)
    assert np.allclose(C @ b, C.full_matrix() @ b)
    assert np.allclose(C.rmatvec(b), b @ C.full_matrix())


def test_even_block_symmetric_circulant_matrix():
    # 5x5 block symmetric circulant matrix representation of the identity matrix
    A = EvenBlockSymmetricCirculantMatrix([
        [np.eye(2, 2), np.zeros((2, 2)), np.zeros((2, 2))]
    ])
    assert A.nb_blocks == (4, 4)
    assert A.shape == (8, 8)
    assert A.block_shapes == ([2, 2, 2, 2], [2, 2, 2, 2])
    assert (A.full_matrix() == np.eye(*A.shape)).all()

    assert ((A + A)/2 == A).all()
    assert (-A).min() == -1
    assert (2*A).max() == 2
    assert (A*A == A).all()

    b = np.random.rand(A.shape[0])
    assert np.allclose(A @ b, A.full_matrix() @ b)
    assert np.allclose(A.rmatvec(b), b @ A.full_matrix())

    B = EvenBlockSymmetricCirculantMatrix([
        [A, A, A]
    ])
    assert B.nb_blocks == (4, 4)
    assert B.shape == (32, 32)
    assert B.all_blocks[0, 0] is A
    assert (B.all_blocks[0, 1] == B.all_blocks[2, 3]).all()

    b = np.random.rand(B.shape[0])
    assert np.allclose(B @ b, B.full_matrix() @ b)
    assert np.allclose(B.rmatvec(b), b @ B.full_matrix())


def test_complex_valued_matrices():
    R = random_block_matrix([2, 2], [2, 2])
    I = random_block_matrix([2, 2], [2, 2])
    A = R + 1j * I
    assert np.allclose(A.full_matrix(), R.full_matrix() + 1j * I.full_matrix())

    B = EvenBlockSymmetricCirculantMatrix([[A.full_matrix(), 2*A.full_matrix()]])
    c = np.random.rand(B.shape[1]) + 1j * np.random.rand(B.shape[1])
    assert np.allclose(B @ c, B.full_matrix() @ c)


def test_low_rank_blocks():
    n = 10

    # Test initialization
    a, b = np.random.rand(n, 1), np.random.rand(1, n)
    LR = LowRankMatrix(a, b)
    assert LR.shape == LR.full_matrix().shape
    assert np.all(np.array(LR) == LR.full_matrix())
    assert matrix_rank(LR.full_matrix()) == LR.rank == 1
    assert LR.density == 2 * n / n ** 2

    a, b = np.random.rand(n, 2), np.random.rand(2, n)
    LR = LowRankMatrix(a, b)
    assert LR.shape == LR.full_matrix().shape
    assert matrix_rank(LR.full_matrix()) == LR.rank == 2

    # Test creation from SVD
    A = np.arange(n**2).reshape((n, n)) + 1.0
    dumb_low_rank = LowRankMatrix.from_full_matrix_with_SVD(A, n)
    assert np.allclose(dumb_low_rank.full_matrix() - A, 0.0)

    A_rank_1 = LowRankMatrix.from_full_matrix_with_SVD(A, 1)
    assert matrix_rank(A_rank_1.full_matrix()) == A_rank_1.rank == 1

    # Test recompression
    recompressed = dumb_low_rank.recompress(new_rank=2)
    assert recompressed.rank == matrix_rank(recompressed.full_matrix()) == 2

    recompressed = dumb_low_rank.recompress(tol=1e-1)
    assert recompressed.rank <= dumb_low_rank.rank

    # Test multiplication with vector
    b = np.random.rand(n)
    assert np.allclose(A_rank_1 @ b, A_rank_1.full_matrix() @ b)

    # Test creation with ACA
    full_A_rank_1 = A_rank_1.full_matrix()
    A_rank_1_again = LowRankMatrix.from_full_matrix_with_ACA(full_A_rank_1, max_rank=5)
    assert matrix_rank(A_rank_1_again.full_matrix()) == 1
    assert np.allclose(A_rank_1_again.full_matrix(), full_A_rank_1)

    # Test creation from function with ACA
    X = np.linspace(0, 1, n)
    Y = np.linspace(10, 11, n)

    def f(i, j):
        return 1/abs(X[i] - Y[j])

    S = np.array([[f(i, j) for j in range(n)] for i in range(n)])
    SLR = LowRankMatrix.from_function_with_ACA(f, n, n, max_rank=2, tol=1e-2)
    assert SLR.shape == (n, n)
    assert np.allclose(SLR.full_matrix(), S, atol=1e-2)

    with pytest.raises(NoConvergenceOfACA):
        LowRankMatrix.from_function_with_ACA(f, n, n, max_rank=1, tol=1e-3)

    summed = SLR + A_rank_1
    assert summed.rank == 1


def test_2_in_1_ACA_with_identical_matrices():
    n = 5
    A = np.fromfunction(lambda i, j: 1/abs(i - (100 + j)), (n, n))
    B = A.copy()

    def get_row_func(i):
        return A[i, :], B[i, :]

    def get_col_func(j):
        return A[:, j], B[:, j]

    lrA, lrB = LowRankMatrix.from_rows_and_cols_functions_with_multi_ACA(
        get_row_func, get_col_func, n, n, nb_matrices=2, tol=1e-3
    )

    assert (lrA.full_matrix() == lrB.full_matrix()).all()
    assert norm(lrA.full_matrix() - A, 'fro')/norm(A, 'fro') < 3e-2


def test_2_in_1_ACA_with_different_matrices():
    n = 5
    A = np.arange(1, 1+n**2).reshape((n, n)) + np.random.rand(n, n)
    B = A.T

    def get_row_func(i):
        return A[i, :], B[i, :]

    def get_col_func(j):
        return A[:, j], B[:, j]

    lrA, lrB = LowRankMatrix.from_rows_and_cols_functions_with_multi_ACA(
        get_row_func, get_col_func, n, n, nb_matrices=2, max_rank=3
    )

    assert matrix_rank(lrA.full_matrix()) == matrix_rank(lrB.full_matrix()) == 3


def test_hierarchical_matrix():
    n = 30
    X = np.linspace(0, 1, n)
    Y = np.linspace(10, 11, n)

    def f(i, j):
        return 1/abs(X[i] - Y[j])

    S = np.array([[f(i, j) for j in range(n)] for i in range(n)])

    HS = BlockMatrix(
        [
            [S[:n//2, :n//2], LowRankMatrix.from_full_matrix_with_ACA(S[n//2:, :n//2], tol=1e-5)],
            [LowRankMatrix.from_full_matrix_with_ACA(S[:n//2, n//2:], tol=1e-5), S[n//2:, n//2:]],
         ]
    )

    assert np.allclose(HS.full_matrix(), S, rtol=2e-1)

    doubled = HS + HS
    assert np.allclose(2*S, doubled.full_matrix(), rtol=2e-1)

def test_rmatvec_lowrank():

    left_matrix = np.arange(6).reshape((3,2))
    right_matrix = np.arange(8).reshape((2,4))

    v = np.ones(3)

    M_lowrank = LowRankMatrix(left_matrix, right_matrix)
    M_full = left_matrix@right_matrix

    vM_lowrank = M_lowrank.__rmatmul__(v)
    vM_full = v @ M_full

    assert np.all(vM_lowrank==vM_full)

def test_access_block_by_path():
    A = BlockMatrix([[random_block_matrix([2, 2], [2, 2]), random_block_matrix([2, 2], [2, 2])],
                     [random_block_matrix([2, 2], [2, 2]), random_block_matrix([2, 2], [2, 2])]])
    assert A.access_block_by_path([0]) is A._stored_blocks[0, 0]
    assert A.access_block_by_path([0, 0]) is A._stored_blocks[0, 0]._stored_blocks[0, 0]
    assert A.access_block_by_path([0, 1]) is A._stored_blocks[0, 0]._stored_blocks[1, 1]
    assert A.access_block_by_path([1, 0]) is A._stored_blocks[1, 1]._stored_blocks[0, 0]



#! /usr/bin/env python3

import numpy as np
import capytaine as cpt


r = 1.
draft = 0.5
depth = 10.
omega = 1.
rho = 1000

# Define geometry and heave degree of freedom
body = cpt.FloatingBody(
        mesh=cpt.mesh_vertical_cylinder(
            length=2*draft, radius=r, center=(0.,0.,0.),
            resolution=(10, 50, 10)
            ))
body.add_translation_dof(name='Heave')
body = body.immersed_part()

solver = cpt.BEMSolver()

# Define and solve the diffraction and radiation problems
diff_problem = cpt.DiffractionProblem(body=body, water_depth=depth,
                                      omega=omega, wave_direction=0.)
rad_problem = cpt.RadiationProblem(body=body, water_depth=depth,
                                   omega=omega, radiating_dof='Heave')

diff_solution = solver.solve(diff_problem)
rad_solution = solver.solve(rad_problem)


# Read mesh properties
faces_centers = body.mesh.faces_centers
faces_normals = body.mesh.faces_normals
faces_areas = body.mesh.faces_areas

from capytaine.bem.airy_waves import airy_waves_potential, airy_waves_velocity, froude_krylov_force

# Computation from the diffraction solution (Capytaine)
FK = froude_krylov_force(diff_problem)['Heave']
diff_force = diff_solution.forces['Heave']

# Get potentials
phi_inc = airy_waves_potential(faces_centers, diff_problem)
v_inc = airy_waves_velocity(faces_centers, diff_problem)
phi_rad = rad_solution.potential

# Indirect computation from the radiation solution, via the Haskind relation
integrand = - (phi_inc * faces_normals[:,2]
              - phi_rad * np.diag(v_inc@faces_normals.T))
F_hask = 1j * omega * rho * np.sum(integrand*faces_areas)

# Direct integration of the potentials
diff_force_recomputed = - 1j * omega * rho * np.sum(
                        diff_solution.potential*faces_areas*faces_normals[:,2])
FK_recomputed = - 1j * omega * rho * np.sum(
                        phi_inc*faces_areas*faces_normals[:,2])

print('Result from Capytaine = ', FK + diff_force)
print('Result from recomputed direct integration',
        FK_recomputed + diff_force_recomputed)
print('Result from Haskind relation = ', F_hask)



#!/usr/bin/env python

import numpy as np
import capytaine as cpt

cpt.set_logging('INFO')

# Initialize floating body by generating a geometric mesh
mesh = cpt.mesh_horizontal_cylinder(
    length=10.0, radius=1.0,  # Dimensions
    center=(0, 0, -2),        # Position
    resolution=(5, 20, 40)    # Fineness of the mesh
    )
body = cpt.FloatingBody(mesh)

# Define a degree of freedom. The keyword "Heave"
# is recognized by the code and the vertical translation
# motion is automatically defined.
body.add_translation_dof(name="Heave")

# Define the range of water depth
depth_range = list(range(5, 25, 2)) + [np.inf]

# Set up the problems: we will solve a radiation problem for each
# water depth:
problems = [
    cpt.RadiationProblem(body=body, water_depth=depth, omega=2.0)
    for depth in depth_range
]
# Water density, gravity and radiating dof have not been specified.
# Default values are used. (For the radiating dof, the default value
# is usually the first one that has been defined. Here only one has
# been defined.)

# Solve all radiation problems
solver = cpt.BEMSolver(engine=cpt.HierarchicalToeplitzMatrixEngine())
results = [solver.solve(pb) for pb in sorted(problems)]

# Gather the computed added mass into a labelled array.
data = cpt.assemble_dataset(results)

# Plot the added mass of each dofs as a function of the water depth.
import matplotlib.pyplot as plt
plt.figure()
plt.plot(
    depth_range,
    data['added_mass'].sel(omega=2.0, radiating_dof="Heave", influenced_dof="Heave"),
    marker="s",
)
plt.xlabel('water depth')
plt.ylabel('added mass')
plt.show()



#!/usr/bin/env python

import numpy as np
import capytaine as cpt

cpt.set_logging()

# Initialize floating body by generating a geometric mesh
cylinder = cpt.FloatingBody(
        mesh=cpt.mesh_horizontal_cylinder(
            length=10.0, radius=1.0,
            center=(0, 0, -2),
            resolution=(10, 20, 30)
            ))

# Automatically add the six degrees of freedom of a rigid body
cylinder.add_all_rigid_body_dofs()

# Define the range of frequencies as a Numpy array
omega_range = np.linspace(0.1, 3.0, 10)

# Set up the problems: we will solve a radiation problem for each
# degree of freedom of the body and for each frequency in the
# frequency range.
problems = [
    cpt.RadiationProblem(body=cylinder, radiating_dof=dof, omega=omega)
    for dof in cylinder.dofs
    for omega in omega_range
]
# Water density, gravity and water depth have not been specified.
# Default values are used.

# Solve all radiation problems
solver = cpt.BEMSolver()
results = solver.solve_all(problems)

# Gather the computed added mass into a labelled array.
data = cpt.assemble_dataset(results)

# Plot the added mass of each dofs as a function of the frequency
import matplotlib.pyplot as plt
plt.figure()
for dof in cylinder.dofs:
    plt.plot(
        omega_range,
        data['added_mass'].sel(radiating_dof=dof, influenced_dof=dof),
        label=dof,
        marker='o',
    )
plt.xlabel('omega')
plt.ylabel('added mass')
plt.legend()
plt.tight_layout()
plt.show()
